// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

#include "vtkCutCell3D.h"

#include "vtkBoundingBox.h"
#include "vtkCell.h"
#include "vtkCellData.h"
#include "vtkCellLocator.h"
#include "vtkDoubleArray.h"
#include "vtkIdList.h"
#include "vtkImplicitPolyDataDistance.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkPolyDataNormals.h"
#include "vtkMathUtilities.h"
#include "vtkMergePoints.h"
#include "vtkObjectFactory.h"
#include "vtkPolyData.h"
#include "vtkPolygon.h"
#include "vtkUnstructuredGrid.h"

VTK_ABI_NAMESPACE_BEGIN
vtkStandardNewMacro(vtkCutCell3D);

enum
{
  UNKNOWN,
  OUTSIDE,
  INSIDE,
  OVERLAP
};

enum
{
  CSHELL,
  CSHELLFACE,
  FACEPGON,
  FACE
};

enum
{
  EXTERIOR,
  INTERIOR,
  BORDER,
  INTBORD
};

enum
{
  ENTRY,
  EXIT,
  TWO,
  CORNER
};

class Cline
{
public:
  Cline()
  {
    p1.fill(0.0);
    p2.fill(0.0);
    line = -1;
  }

  // coords of end points of line clipped to cell
  std::array<double, 3> p1;
  std::array<double, 3> p2;
  // index in list of lines that intersect this cell
  int line;
};

class Point
{
public:
  Point()
  {
    x.fill(0.0);
    type = ENTRY;
    next = EXIT;
    line = ENTRY;
    corner = -1;
    cprev = cnext = -1;
    side = -1;
    value = 0.0;
  }

  std::array<double, 3> x; // coords of point
  int type; // typef of pt = ENTRY,EXIT,TWO,CORNER
    // ENTRY/EXIT = only part of one Cline,
    //   could also be a geometric corner pt
    // TWO = part of two CLines
    // CORNER = part of no CLines, is a geometric corner pt
  int next; // index of next point when walking a loop
    // set for ENTRY/TWO pts between ENTRY and EXIT
    // set for EXIT/CORNER points around cell perimerter,
    //   though may not be walked
  int line; // original line (as stored by Cline) the pt starts,
    //   only set for ENTRY and TWO pts
  int corner; // 1,2,3,4 if x is a corner point, else 0
    // could be ENTRY,EXIT,CORNER pt, but not a TWO pt
  int cprev, cnext; // indices of pts in linked list around cell perimeter
  int side; // which side of cell (0,1,2,3) pt is on
    // only for ENTRY/EXIT/CORNER pts to make linked list
  double value; // coord along the side
    // only for ENTRY/EXIT/CORNER pts to make linked list
};

class Loop2D
{
public:
  Loop2D()
  {
    area = 0.0;
    active = 0;
    flag = INTERIOR;
    n = 0;
    first = -1;
    next = -1;
  }

  double area; // area of loop
  int active; // 1/0 if active or not
  int flag; // INTERIOR (if all TWO points) or BORDER
  vtkIdType n; // # of points in loop
  vtkIdType first; // index of first point in loop
  vtkIdType next; // index of loop in same Polygon, -1 if last loop
};

class Loop3D
{
public:
  Loop3D()
  {
    volume = 0.0;
    flag = INTERIOR;
    n = 0;
    first = -1;
    next = -1;
  }

  double volume; // volume of loop
  int flag; // INTERIOR (if all CSHELL vertices) or BORDER
  vtkIdType n; // # of vertices in loop
  vtkIdType first; // index of first vertex in loop
  vtkIdType next; // index of next loop in same Polyhedron, -1 if last loop
};

class Polygon
{
public:
  Polygon()
  {
    area = 0.0;
    n = -1;
    first = -1;
  }

  double area; // summed area (over loops) of Polygon
  vtkIdType n; // # of loops in Polygon
  vtkIdType first; // index of first loop in Polygon
};

class Polyhedron
{
public:
  Polyhedron()
  {
    volume = 0.0;
    n = -1;
    first = -1;
  }

  double volume;
  vtkIdType n;
  vtkIdType first;
};

class CutCellInfo
{
public:
  CutCellInfo()
  {
    type = UNKNOWN;
    corner.fill(UNKNOWN);
    volume = 0.0;
  }

  int type; // OUTSIDE,INSIDE,OVERLAP,UNKNOWN
  std::array<int, 8> corner; // conrer flag
    // OUTSIDE,INSIDE,UNKNOWN
    // ordered x first, y next, z last
    // for sub cells, type/coner are smae as input cell containing them
  double volume; // flow volume of cell or sub cell
    // entire cell volume for split cell
};

class Vertex
{
public:
  Vertex()
  {
    active = 0;
    style = CSHELL;
    label = -1;
    next = -1;
    nEdge = 0;
    first = -1;
    dirFirst = -1;
    last = -1;
    dirLast = -1;
    volume = 0.0;
    norm.fill(0.0);
  }

  int active; // 1/0 if active or not
  int style; // CSHELL or CSHELLFACE or FACEPGON or FACE
  int label; // index in list of shells that intersect this cell
    // face index (0-5) for FACEPGON or FACE
  vtkIdType next; // index of next vertex when walking a loop
  vtkIdType nEdge; // # of edges in this vertex
  vtkIdType first; // first edge in vertex
  int dirFirst; // dir of fist edge in vertex
  vtkIdType last; // last edge in vertex
  int dirLast; // dir of last edge in vertex
  double volume; // volume of vertex projected against lower z face of cell
  std::array<double, 3> norm; // norm of shell, NULL for other styles
};

class Edge
{
public:
  Edge()
  {
    active = 0;
    style = CSHELL;
    nVert = 0;
    p1.fill(0.0);
    p2.fill(0.0);
    verts.fill(-1);
    next.fill(-1);
    dirNext.fill(-1);
    prev.fill(-1);
    dirPrev.fill(-1);
  }

  std::array<double, 3> p1;
  std::array<double, 3> p2;
  int active; // 1/0 if active or not
  int style; // CSHELL or CSHELLFACE or FACEPGON or FACE
  int clipped; // 1/0 if already clipped during face iteration
  int nVert; // flag for verts containing this edge
    // 0 = no verts
    // 1 = just 1 vert in forward dir
    // 2 = just 1 vert in reverse dir
    // 3 = 2 verts in both dirs
    // all vecs are [0] in forward dir, [1] in reverse dir
  std::array<vtkIdType, 2> verts; // index of vertices containing this edge, -1 if not
  std::array<vtkIdType, 2> next; // index of next edge for each vertex, -1 for end
  std::array<int, 2> dirNext; // next edge for each vertex is forward/reverse (0,1)
  std::array<vtkIdType, 2> prev; // index of prev edge for each vertex, -1 for start
  std::array<int, 2> dirPrev; // prev edge for each vertex is foward/reverse (0,1)
};

//-----------------------------------------------------------------------------
vtkCutCell3D::vtkCutCell3D()
  : UseInputBounds(true)
  , InsideOut(false)
  , ComputeVolume(false)
  , VolumeArrayName(nullptr)
{
  this->Output = nullptr;
  this->Source = nullptr;
  this->Points = nullptr;
  this->Locator = nullptr;
  this->CellLocator = nullptr;
  this->ImplicitFunction = nullptr;
  this->Dimension = 3;
  this->Tolerance = 1.0e-14;
  this->EpsilonEdge = 1.0e-9;
  this->Shrink = 0.0;
  this->SetVolumeArrayName("Volume");

  this->CellSurfs = vtkIdList::New();

  this->GridBounds[0] = this->GridBounds[2] = this->GridBounds[4] = 0;
  this->GridBounds[1] = this->GridBounds[3] = this->GridBounds[5] = 1;
  this->GridDimensions[0] = this->GridDimensions[1] = this->GridDimensions[2] = 10;

  this->Path1 = new double* [VTK_CELL_SIZE];
  this->Path2 = new double* [VTK_CELL_SIZE];
  for (int i = 0; i < VTK_CELL_SIZE; ++i)
  {
    this->Path1[i] = new double[3];
    this->Path2[i] = new double[3];
  }

  this->SetNumberOfInputPorts(1);
}

//-----------------------------------------------------------------------------
vtkCutCell3D::~vtkCutCell3D()
{
  this->Points->Delete();
  this->Locator->Delete();
  this->CellLocator->Delete();
  this->CellSurfs->Delete();

  this->Points = nullptr;
  this->Locator = nullptr;
  this->CellLocator = nullptr;
  this->CellSurfs = nullptr;
  this->SetVolumeArrayName(nullptr);
  this->ImplicitFunction = nullptr;

  for (int i = 0; i < VTK_CELL_SIZE; ++i)
  {
    delete[] this->Path1[i];
    delete[] this->Path2[i];
  }
  delete this->Path1;
  delete this->Path2;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);
  os << indent << "UseInputBounds " << this->UseInputBounds << endl;
  os << indent << "GridBounds [" << this->GridBounds[0] << ", " << this->GridBounds[1]
    << ", " << this->GridBounds[2] << ", " << this->GridBounds[3] << ", "
    << this->GridBounds[4] << ", " << this->GridBounds[5] << "]" << endl;
  os << indent << "GridDimensions " << this->GridDimensions[0] << " x "
    << this->GridDimensions[1] << " x " << this->GridDimensions[2] << std::endl;
  os << indent << "Tolerance: " << this->Tolerance << std::endl;
  os << indent << "Shrink: " << this->Shrink << std::endl;
  os << indent << "ComputeVolume: " << this->ComputeVolume << std::endl;
  if (this->VolumeArrayName)
  {
    os << indent << "VolumeArrayName: " << this->VolumeArrayName << std::endl;
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::FillInputPortInformation(int, vtkInformation* info)
{
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkPolyData");
  return 1;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::SetLocator(vtkIncrementalPointLocator* locator)
{
  if (this->Locator == locator)
  {
    return;
  }
  if (this->Locator)
  {
    this->Locator->UnRegister(this);
    this->Locator = nullptr;
  }

  if (locator)
  {
    locator->Register(this);
  }

  this->Locator = locator;
  this->Locator->SetTolerance(this->Tolerance);
  this->Modified();
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::CreateDefaultLocator()
{
  if (this->Locator == nullptr)
  {
    this->Locator = vtkMergePoints::New();
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::RequestData(vtkInformation* vtkNotUsed(request),
  vtkInformationVector** inputVector, vtkInformationVector* outputVector)
{
  // Get the info objects
  vtkInformation* inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation* outInfo = outputVector->GetInformationObject(0);

  // Get the input and output
  this->Source = vtkPolyData::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
  this->Output =
    vtkUnstructuredGrid::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkNew<vtkPolyDataNormals> normalsFilter;
  normalsFilter->SetInputData(this->Source);
  normalsFilter->ConsistencyOn();
  normalsFilter->SplittingOff();
  if (this->InsideOut)
  {
    normalsFilter->FlipNormalsOn();
  }
  normalsFilter->Update();
  this->Source->ShallowCopy(normalsFilter->GetOutput());

  double gridBounds[6];
  if (this->UseInputBounds)
  {
    this->Source->GetBounds(gridBounds);

    // To avoid accidentally sampling outside the dataset due to floating point roundoff,
    // nudge the bounds inward by epsilon.
    vtkBoundingBox bbox(gridBounds);
    const double eps = 1.0e-6;
    bbox.ScaleAboutCenter(1.0 - eps);
    bbox.GetBounds(gridBounds);
  }
  else
  {
    std::copy(this->GridBounds, this->GridBounds + 6, gridBounds);
  }

  // bounds and extent where cut cell should be performed
  double origin[3] = { gridBounds[0], gridBounds[2], gridBounds[4] };
  double spacing[3];
  int numGridCell = GridDimensions[0] * GridDimensions[1] * GridDimensions[2];
  for (int i = 0; i < 3; ++i)
  {
    spacing[i] = (this->GridDimensions[i] == 1)
      ? 0
      : ((gridBounds[i * 2 + 1] - gridBounds[i * 2]) /
        static_cast<double>(this->GridDimensions[i] - 1));
  }

  int extent[6];
  for (int i = 0; i < 3; ++i)
  {
    if (spacing[i] != 0.0)
    {
      extent[2 * i] = static_cast<int>(vtkMath::Floor((gridBounds[2 * i] - origin[i]) / spacing[i]));
      extent[2 * i + 1] =
        static_cast<int>(vtkMath::Ceil((gridBounds[2 * i + 1] - origin[i]) / spacing[i]));
    }
    else
    {
      extent[2 * i] = extent[2 * i + 1] = 0;
    }
  }

  this->Output->Allocate(numGridCell);

  this->Points = vtkPoints::New();
  this->CellLocator = vtkCellLocator::New();
  this->CellLocator->SetDataSet(this->Source);
  this->CellLocator->BuildLocator();

  // Initialize output cell data
  this->OutData = this->Output->GetCellData();
  this->OutData->Allocate(numGridCell);

  if (this->ComputeVolume)
  {
    auto array = vtkSmartPointer<vtkDoubleArray>::New();
    array->SetName(this->VolumeArrayName);
    array->Allocate(numGridCell);
    this->OutData->AddArray(array);
  }

  // Locator used  to merge potentially duplicate points
  if (this->Locator == nullptr)
  {
    this->CreateDefaultLocator();
  }
  this->Locator->InitPointInsertion(this->Points, gridBounds);

  vtkNew<vtkImplicitPolyDataDistance> implicitFunction;
  implicitFunction->SetInput(this->Source);
  this->ImplicitFunction = implicitFunction;

  // Iterate over all grid cell
  double cellOrigin[3];
  for (int k = extent[4]; k < extent[5]; ++k)
  {
    cellOrigin[2] = origin[2] + k * spacing[2];
    for (int j = extent[2]; j < extent[3]; ++j)
    {
      cellOrigin[1] = origin[1] + j * spacing[1];
      for (int i = extent[0]; i < extent[1]; ++i)
      {
        vtkIdType cellId = i + j * GridDimensions[0]
          + k * GridDimensions[0] * GridDimensions[1];
        cellOrigin[0] = origin[0] + i * spacing[0];

        // Cut cell
        this->CutCell(cellId, cellOrigin, spacing);

        if (this->ComputeVolume)
        {
          for (auto& vol : this->Vols)
          {
            this->OutData->GetArray(this->VolumeArrayName)->InsertNextTuple1(vol);
          }
        }
      }
    }
  }

  this->Output->SetPoints(this->Points);

  return 1;
}

//----------------------------------------------------------------------------
void vtkCutCell3D::AddHex(vtkIdType cellId, double* origin, double* size)
{
  // Storage for point coordinates
  double pt[3] = {0.0, 0.0, 0.0};

  // Storage for cell vertex IDs
  vtkIdType ids[8];

  double volume = 0.0;

  // First cell vertex is always at origin of cursor
  memcpy(pt, origin, 3 * sizeof(double));
  this->Locator->InsertUniquePoint(pt, ids[0]);

  // Create remaining 2^d - 1 vertices depending on dimension
  switch (this->Dimension)
  {
  case 1:
    // In 1D there is only one other vertex
    pt[0] = origin[0] + size[0];
    this->Locator->InsertUniquePoint(pt, ids[1]);

    // Insert next line
    this->Output->InsertNextCell(VTK_LINE, 2, ids);
    volume = 0.0;
    break;
  case 2:
    // Add vertex #1 : (1,0)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1];
    this->Locator->InsertUniquePoint(pt, ids[1]);

    // Add vertex #2 : (0,1)
    pt[0] = origin[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[2]);

    // Add vertex #3 : (1,1)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[3]);

    // Insert next quadrangle
    this->Output->InsertNextCell(VTK_QUAD, 4, ids);
    volume = 0.0;
    break;
  case 3:
    // z=0 plane
    pt[2] = origin[2];

    // Add vertex #1 : (1,0,0)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1];
    this->Locator->InsertUniquePoint(pt, ids[1]);

    // Add vertex #2 : (1,1,0)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[2]);

    // Add vertex #3 : (0,1,0)
    pt[0] = origin[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[3]);

    // z=1 plane
    pt[2] = origin[2] + size[2];

    // Add vertex #4 : (0,0,1)
    pt[0] = origin[0];
    pt[1] = origin[1];
    this->Locator->InsertUniquePoint(pt, ids[4]);

    // Add vertex #5 : (1,0,1)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1];
    this->Locator->InsertUniquePoint(pt, ids[5]);

    // Add vertex #6 : (1,1,1)
    pt[0] = origin[0] + size[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[6]);

    // Add vertex #7 : (0,1,1)
    pt[0] = origin[0];
    pt[1] = origin[1] + size[1];
    this->Locator->InsertUniquePoint(pt, ids[7]);

    // Insert next hexahedron
    this->Output->InsertNextCell(VTK_HEXAHEDRON, 8, ids);
    volume = size[0] * size[1] * size[2];
    break;
  default:
    return;
  } // switch (this->Dimension)

  // store volumes in vector
  this->Vols.resize(1);
  this->Vols[0] = volume;
}

//----------------------------------------------------------------------------
void vtkCutCell3D::AddPolyhedron(vtkIdType cellId,
  double* lo, double* hi, vtkUnstructuredGrid* output)
{
  this->Vols.clear();

  // Storage for point coordinates
  double* p1, * p2;

  int iVert, firstVert, iEdge, dir, nEdge;
  Vertex* vert;
  Edge* edge;

  // used = 0/1 flag for whether a vertex is already part of a loop
  // only active vertices are eligible
  int nVert = this->nVerts;
  this->Used.clear();
  this->Used.resize(nVert, 1);
  for (int iVert = 0; iVert < nVert; ++iVert)
  {
    if (this->Verts[iVert].active)
    {
      this->Used[iVert] = 0;
    }
  }

  // stack = list of vertex indices to process
  // max size = # of vertices
  this->Stack.clear();
  this->Stack.resize(nVert, -1);
  int nStack = 0;

  // Iterate over all vertices
  // start a loop at any unused vertex
  // add more vertices to loop via stack
  // check all neighbor vertices via shared edges
  // if neighbor vertex is unused, add to stack
  // stop when stack is empty
  for (int i = 0; i < nVert; ++i)
  {
    if (this->Used[i])
    {
      continue;
    }

    double volume = 0.0;
    int nfaces = 0;

    this->Stack[0] = firstVert = i;
    nStack = 1;
    this->Used[i] = 1;
    int prev = -1;

    std::vector<vtkIdType> faceIds;

    while (nStack)
    {
      --nStack;
      iVert = this->Stack[nStack];
      ++nfaces;

      vert = &this->Verts[iVert];
      volume += vert->volume;

      nEdge = vert->nEdge;
      iEdge = vert->first;
      dir = vert->dirFirst;

      std::vector<vtkIdType> tempFaces;
      tempFaces.reserve(2 * nEdge);
      for (int i = 0; i < nEdge; ++i)
      {
        edge = &this->Edges[iEdge];
        if (!this->Used[edge->verts[0]])
        {
          this->Stack[nStack++] = edge->verts[0];
          this->Used[edge->verts[0]] = 1;
        }
        if (!this->Used[edge->verts[1]])
        {
          this->Stack[nStack++] = edge->verts[1];
          this->Used[edge->verts[1]] = 1;
        }

        p1 = edge->p1.data();
        p2 = edge->p2.data();

        vtkIdType ptId;
        this->Locator->InsertUniquePoint(p1, ptId);
        tempFaces.emplace_back(ptId);
        this->Locator->InsertUniquePoint(p2, ptId);
        tempFaces.emplace_back(ptId);

        iEdge = edge->next[dir];
        dir = edge->dirNext[dir];
      }

      // Rearrange the edge endpoints Ids so that they are connected.
      for (int j = 1; j < nEdge; ++j)
      {
        int preId1 = tempFaces[2 * (j - 1)];
        int preId2 = tempFaces[2 * (j - 1) + 1];
        int id1 = tempFaces[2 * j];
        int id2 = tempFaces[2 * j + 1];

        if (preId2 != id1 && preId2 != id2)
        {
          std::swap(tempFaces[2 * (j - 1)], tempFaces[2 * (j - 1) + 1]);
          std::swap(preId1, preId2);
        }

        if (preId2 != id1)
        {
          std::swap(tempFaces[2 * j], tempFaces[2 * j + 1]);
        }
      }

      faceIds.emplace_back(nEdge);
      for (int j = 0; j < nEdge; ++j)
      {
        faceIds.emplace_back(tempFaces[2 * j]);
      }

      if (prev >= 0)
      {
        this->Verts[prev].next = iVert;
      }
      prev = iVert;
    }
    vert->next = -1;

    vtkIdType outId = output->InsertNextCell(VTK_POLYHEDRON, nfaces, faceIds.data());

    // store volumes in vector
    this->Vols.emplace_back(volume);
  }
}

//----------------------------------------------------------------------------
void vtkCutCell3D::CutCell(vtkIdType cellId, double* origin, double* size)
{
  int xSub;
  double xSplit[3];

  this->CellSurfs->Reset();
  this->Vols.clear();

  double sBounds[6];
  this->Source->GetBounds(sBounds);
  vtkBoundingBox surfBBox(sBounds);

  double lo[3], hi[3];
  for (unsigned d = 0; d < this->Dimension; ++d)
  {
    lo[d] = origin[d];
    hi[d] = origin[d] + size[d];
  }
  vtkBoundingBox cellBBox(lo[0], hi[0], lo[1], hi[1], lo[2], hi[2]);

  // Iterate over all vertices
  int nPos = 0;
  int nNeg = 0;
  int nVert = 1 << this->Dimension;
  for (int v = 0; v < nVert; ++v)
  {
    // Transform flat index into triple
    div_t d1 = div(v, 2);
    div_t d2 = div(d1.quot, 2);

    // Compute vertex coordinates
    double pt[3];
    pt[0] = origin[0] + d1.rem * size[0];
    pt[1] = origin[1] + d2.rem * size[1];
    pt[2] = origin[2] + d2.quot * size[2];

    // Evalute implicit value at current vertex
    double qv = this->ImplicitFunction->EvaluateFunction(pt);
    if (qv > 0)
    {
      // Found positive value at this vertex
      ++nPos;
    }
    else if (qv < 0)
    {
      ++nNeg;
    }
  }

  if (nNeg == nVert)
  {
    int isIntersect = 0;
    for (vtkIdType ptId = 0; ptId < this->Source->GetNumberOfPoints(); ++ptId)
    {
      if (cellBBox.ContainsPoint(this->Source->GetPoint(ptId)))
      {
        isIntersect = 1;
        break;
      }
    }
    if (!isIntersect)
    {
      return;
    }
  }

  if (!this->SurfToGrid(cellId, cellBBox))
  {
    this->AddHex(cellId, origin, size);
    return;
  }

  // shrinking grid cell by Shrink factor
  double newLo[3] = { lo[0] + this->Shrink * (hi[0] - lo[0]),
    lo[1] + this->Shrink * (hi[1] - lo[1]),
    lo[2] + this->Shrink * (hi[2] - lo[2]) };
  double newHi[3] = { hi[0] - this->Shrink * (hi[0] - lo[0]),
    hi[1] - this->Shrink * (hi[1] - lo[1]),
    hi[2] - this->Shrink * (hi[2] - lo[2]) };

  this->CellInfo[cellId].type = OVERLAP;
  if (this->Split(cellId, newLo, newHi, xSub, xSplit) == -1)
  {
    this->AddHex(cellId, origin, size);
    return;
  }
  this->AddPolyhedron(cellId, newLo, newHi, this->Output);
}

//----------------------------------------------------------------------------
int vtkCutCell3D::SurfToGrid(vtkIdType cellId, const vtkBoundingBox& cellBBox)
{
  double shellBounds[6], cellBounds[6];
  cellBBox.GetBounds(cellBounds);

  vtkPoints* points = this->Source->GetPoints();

  auto cells = vtkSmartPointer<vtkIdList>::New();
  auto ptIds = vtkSmartPointer<vtkIdList>::New();
  this->CellLocator->FindCellsWithinBounds(cellBounds, cells);

  this->nSurf = 0;
  for (vtkIdType i = 0; i < cells->GetNumberOfIds(); ++i)
  {
    int shellId = cells->GetId(i);
    this->Source->GetCell(shellId)->GetBounds(shellBounds);
    vtkBoundingBox shellBBox(shellBounds);
    this->Source->GetCellPoints(shellId, ptIds);

    // If there is no intersection, nothing is done
    if (!shellBBox.IntersectBox(cellBBox))
    {
      continue;
    }

    // Sutherland-Hodgman clip algorithm
    if (this->Clip(shellId, points, shellBBox, cellBBox))
    {
      ++this->nSurf;
      this->CellSurfs->InsertNextId(shellId);
    }
  }

  return this->nSurf;
}

//----------------------------------------------------------------------------
int vtkCutCell3D::Split(vtkIdType cellId, double* lo, double* hi, int& xSub, double* xSplit)
{
  int errFlag;
  int nSplit;

  this->Verts.clear();
  this->Edges.clear();

  while (1)
  {
    errFlag = this->AddShells(cellId);
    if (errFlag)
    {
      break;
    }

    this->ClipShells(cellId, lo, hi);
    this->ClipAdjust(cellId, lo, hi);

    // all shells just touched cell surface
    // mark corner points based on GrazeCount or TouchMark value
    // return vol = 0.0 for UNKNOWN/INSIDE, full cell vol for OUTSIDE
    // vol is changed in SetInOut() if OVERLAP cell corners are marked
    if (this->Empty)
    {
      int mark = UNKNOWN;
      if (this->GrazeCount || this->TouchMark == INSIDE)
      {
        mark = INSIDE;
      }
      else if (this->TouchMark == OUTSIDE)
      {
        mark = OUTSIDE;
      }
      this->CellInfo[cellId].corner.fill(mark);

      double vol = 0.0;
      if (mark == OUTSIDE)
      {
        vol = (hi[0] - lo[0]) * (hi[1] - lo[1]) * (hi[2] - lo[2]);
      }

      this->Vols.emplace_back(vol);
      return -1;
    }

    this->CellShellVolume(lo, hi);
    errFlag = this->EdgeToFace(lo, hi);
    if (errFlag)
    {
      break;
    }

    double lo2d[2], hi2d[2];
    for (int iFace = 0; iFace < 6; ++iFace)
    {
      this->FaceFromCell(iFace, lo, hi, lo2d, hi2d);

      if (this->FaceList[iFace].size())
      {
        this->EdgeToCLines(iFace);

        errFlag = this->SplitFace(cellId, iFace, lo2d, hi2d);
        if (errFlag)
        {
          break;
        }
        errFlag = this->AddFacePolygons(iFace, lo, hi);
        if (errFlag)
        {
          break;
        }
      }
      else
      {
        errFlag = this->AddFace(iFace, lo, hi, lo2d, hi2d);
        if (errFlag)
        {
          break;
        }
      }
    }

    if (errFlag)
    {
      break;
    }

    this->RemoveFaces();

    errFlag = this->Check();
    if (errFlag)
    {
      break;
    }

    this->Walk();
    errFlag = this->LoopToPolyhedron();

    // LoopToPolyhedron detected no positive-volume loops, cell is inside the surf
    if (errFlag == 4)
    {
      this->CellInfo[cellId].corner.fill(INSIDE);
      double volume = 0.0;
      this->Vols.resize(1);
      this->Vols[0] = volume;
      nSplit = 1;
      return 0;
    }

    // other error returns from LoopToPolyhedron
    if (errFlag)
    {
      break;
    }

    nSplit = this->nPolyhedrons;
    if (nSplit > 1)
    {
      this->CreateSurfMap();
      errFlag = this->SplitPoint(cellId, lo, hi, xSplit, xSub);
      if (errFlag)
      {
        break;
      }
    }

    // successful cut/split
    // set corners = OUTSIDE if corner pt is in list of edge points
    // else set corners = INSIDE
    int iCorner;
    double* p1, * p2;

    this->CellInfo[cellId].corner.fill(INSIDE);
    int nEdge = this->nEdges;
    for (int iEdge = 0; iEdge < nEdge; ++iEdge)
    {
      if (!this->Edges[iEdge].active)
      {
        continue;
      }

      p1 = this->Edges[iEdge].p1.data();
      p2 = this->Edges[iEdge].p2.data();
      iCorner = this->Corner(p1, lo, hi);
      if (iCorner >= 0)
      {
        this->CellInfo[cellId].corner[iCorner] = OUTSIDE;
      }
      iCorner = this->Corner(p2, lo, hi);
      if (iCorner >= 0)
      {
        this->CellInfo[cellId].corner[iCorner] = OUTSIDE;
      }
    }

    break;
  }

  // could not perform cut/split -> fatal error
  // print into about cell and final error message
  if (errFlag)
  {
    if (errFlag == 1)
    {
      vtkErrorMacro(<< "FindEdge(): Found edge in same direction");
    }
    if (errFlag == 2)
    {
      vtkErrorMacro(<< "EdgeToFace(): Singlet BPG edge not on cell face");
    }
    if (errFlag == 3)
    {
      vtkErrorMacro(<< "EdgeToFace(): BPG edge on more than 2 faces");
    }
    if (errFlag == 4)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): No positive volumes in cell");
    }
    if (errFlag == 5)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): More than one positive volume with "
        << "a negative volume");
    }
    if (errFlag == 6)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): Single volume is negative, inverse donut");
    }
    if (errFlag == 7)
    {
      vtkErrorMacro(<< "SplitPoint(): Could not find split point in split cell");
    }

    if (errFlag == 11)
    {
      vtkErrorMacro(<< "Check(): Vertex has less than 3 edges");
    }
    if (errFlag == 12)
    {
      vtkErrorMacro(<< "Check(): Vertex contains invalid edge");
    }
    if (errFlag == 13)
    {
      vtkErrorMacro(<< "Check(): Vertex contains edge that doesn't point to it");
    }
    if (errFlag == 14)
    {
      vtkErrorMacro(<< "Check(): Vertex contains duplicate edge");
    }
    if (errFlag == 15)
    {
      vtkErrorMacro(<< "Check(): Vertex pointers to last edge are invalid");
    }
    if (errFlag == 16)
    {
      vtkErrorMacro(<< "Check(): Edge not part of 2 vertices");
    }
    if (errFlag == 17)
    {
      vtkErrorMacro(<< "Check(): Edge part of same vertex twice");
    }
    if (errFlag == 18)
    {
      vtkErrorMacro(<< "Check(): Edge part of invalid vertex");
    }
    if (errFlag == 19)
    {
      vtkErrorMacro(<< "Check(): Edge part of invalid vertex");
    }

    if (errFlag == 21)
    {
      vtkErrorMacro(<< "WeilerBuild(): Point appears first in more than one CLINE");
    }
    if (errFlag == 22)
    {
      vtkErrorMacro(<< "WeilerBuild(): Point appears last in more than one CLINE");
    }
    if (errFlag == 23)
    {
      vtkErrorMacro(<< "WeilerBuild(): Singlet CLINES point not on cell border");
    }
    if (errFlag == 24)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): No positive areas in cell");
    }
    if (errFlag == 25)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): More than one positive area with a negative area");
    }
    if (errFlag == 26)
    {
      vtkErrorMacro(<< "LoopToPolyhedron(): Single area is negative, inverse donut");
    }
  }

  return nSplit;
}

//----------------------------------------------------------------------------
int vtkCutCell3D::SplitPoint(vtkIdType cellId, double* lo, double* hi, double* xSplit, int& xSub)
{
  vtkBoundingBox cellBBox(lo[0], hi[0], lo[1], hi[1], lo[2], hi[2]);

  vtkPoints* points = this->Source->GetPoints();
  vtkIdList* idList = vtkIdList::New();
  vtkIdType nSurf = this->CellSurfs->GetNumberOfIds();

  // if end pt of any line with non-negative surfmap is in/on cell, return
  for (vtkIdType i = 0; i < nSurf; ++i)
  {
    if (this->SurfMap[i] < 0)
    {
      continue;
    }
    vtkIdType shellId = this->CellSurfs->GetId(i);
    vtkCell* shell = this->Source->GetCell(shellId);
    idList = shell->GetPointIds();

    for (vtkIdType j = 0; j < idList->GetNumberOfIds(); ++j)
    {
      double pt[3];
      points->GetPoint(idList->GetId(j), pt);
      if (this->PtFlag(pt, lo, hi) != EXTERIOR)
      {
        memcpy(xSplit, pt, 3 * sizeof(double));
        xSub = this->SurfMap[i];
        return 0;
      }
    }
  }

  // clip 1st line with non-negative surfMap to cell, and return clip point
  for (vtkIdType i = 0; i < nSurf; ++i)
  {
    if (this->SurfMap[i] < 0)
    {
      continue;
    }
    int shellId = this->CellSurfs->GetId(i);
    vtkCell* shell = this->Source->GetCell(shellId);

    vtkBoundingBox shellBBox(shell->GetBounds());
    this->Clip(shellId, points, shellBBox, cellBBox);
    memcpy(xSplit, this->Path1[0], 3 * sizeof(double));
    xSub = this->SurfMap[i];

    return 0;
  }

  return 7;
}

//----------------------------------------------------------------------------
int vtkCutCell3D::AddShells(vtkIdType cellId)
{
  double p1[3], p2[3];
  int e, dir;
  Edge* edge;

  vtkIdList* idList = vtkIdList::New();
  vtkPoints* points = this->Source->GetPoints();
  vtkIdType nSurf = this->CellSurfs->GetNumberOfIds();

  this->Verts.resize(nSurf);
  this->Edges.reserve(3 * nSurf);
  this->nVerts = 0;
  this->nEdges = 0;

  int nVert = 0;
  for (vtkIdType i = 0; i < nSurf; ++i, ++nVert)
  {
    vtkIdType shellId = this->CellSurfs->GetId(i);
    idList = this->Source->GetCell(shellId)->GetPointIds();
    int ptNum = idList->GetNumberOfIds();

    if (ptNum < 3)
    {
      continue;
    }

    Vertex& vert = this->Verts[nVert];
    vert.active = 1;
    vert.label = i;
    vtkPolygon::ComputeNormal(points, ptNum, idList->GetPointer(0), vert.norm.data());

    // Look for each edge of shell.
    // Add to edges in forward dir if doesn't yet exist.
    // Add to edges in returned dir if already exists.
    int preE = -1;
    int preDir = -1;
    for (vtkIdType ptId = 0; ptId < ptNum; ++ptId)
    {
      points->GetPoint(idList->GetId(ptId), p1);
      points->GetPoint(idList->GetId((ptId + 1) % ptNum), p2);

      e = this->FindEdge(p1, p2, 0, dir);
      if (e == -2)
      {
        return 1;
      }

      if (e < 0)
      {
        this->Edges.push_back(Edge());
        e = this->nEdges++;
        dir = 0;
        edge = &this->Edges.back();
        memcpy(edge->p1.data(), p1, 3 * sizeof(double));
        memcpy(edge->p2.data(), p2, 3 * sizeof(double));
      }

      this->EdgeInsert(e, dir, nVert, preE, preDir, -1, -1);
      preE = e;
      preDir = dir;
    }
  }
  this->nVerts = nVert;

  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::ClipShells(vtkIdType cellId, double* lo, double* hi)
{
  vtkIdType nEdge, iEdge, jEdge;
  int iDir, jDir;
  int p1Flag, p2Flag;
  double* p1, * p2;
  Edge* edge, * newEdge;

  // loop over all 6 faces of cell
  vtkIdType nVert = this->nVerts;

  for (int iFace = 0; iFace < 6; ++iFace)
  {
    int dim = iFace / 2;
    int lohi = iFace % 2;
    double value = (lohi == 0) ? lo[dim] : hi[dim];

    // mark all edges as unclipped
    // some may have been clipped and not cleared on previous face
    for (auto& edge : this->Edges)
    {
      if (edge.active)
      {
        edge.clipped = 0;
      }
    }

    // loop over vertices, clip each of its edges to face
    // if edge already clipped, unset clip flag and keep edge as-is
    vtkIdType iVert = 0;
    for (auto& vert : Verts)
    {
      iEdge = vert.first;
      iDir = vert.dirFirst;
      nEdge = vert.nEdge;

      for (vtkIdType i = 0; i < nEdge; ++i)
      {
        edge = &this->Edges[iEdge];

        if (edge->clipped)
        {
          edge->clipped = 0;
          iEdge = edge->next[iDir];
          iDir = edge->dirNext[iDir];
          continue;
        }

        // p1/p2 are pts in order of traversal
        p1 = edge->p1.data();
        p2 = edge->p2.data();
        if (iDir != 0)
        {
          std::swap(p1, p2);
        }

        // p1/p2 flag = OUTSIDE/ON/INSIDE for edge pts
        if (lohi == 0)
        {
          if (p1[dim] < value)
          {
            p1Flag = OUTSIDE;
          }
          else if (p1[dim] > value)
          {
            p1Flag = INSIDE;
          }
          else
          {
            p1Flag = OVERLAP;
          }
          if (p2[dim] < value)
          {
            p2Flag = OUTSIDE;
          }
          else if (p2[dim] > value)
          {
            p2Flag = INSIDE;
          }
          else
          {
            p2Flag = OVERLAP;
          }
        }
        else
        {
          if (p1[dim] < value)
          {
            p1Flag = INSIDE;
          }
          else if (p1[dim] > value)
          {
            p1Flag = OUTSIDE;
          }
          else
          {
            p1Flag = OVERLAP;
          }
          if (p2[dim] < value)
          {
            p2Flag = INSIDE;
          }
          else if (p2[dim] > value)
          {
            p2Flag = OUTSIDE;
          }
          else
          {
            p2Flag = OVERLAP;
          }
        }

        // if both OUTSIDE or one OUTSIDE and other ON, delete edge
        // if both INSIDE or one INSIDE and other ON or both ON, keep as-is
        // if one INSIDE and one OUTSIDE, replace OUTSIDE pt with clip pt
        if (p1Flag == OUTSIDE)
        {
          if (p2Flag == OUTSIDE || p2Flag == OVERLAP)
          {
            this->EdgeRemove(edge, iDir);
          }
          else
          {
            if (iDir == 0)
            {
              this->Between(p1, p2, dim, value, edge->p1.data());
            }
            else
            {
              this->Between(p1, p2, dim, value, edge->p2.data());
            }
            edge->clipped = 1;
          }
        }
        else if (p1Flag == INSIDE)
        {
          if (p2Flag == OUTSIDE)
          {
            if (iDir == 0)
            {
              this->Between(p1, p2, dim, value, edge->p2.data());
            }
            else
            {
              this->Between(p1, p2, dim, value, edge->p1.data());
            }
            edge->clipped = 1;
          }
        }
        else
        {
          if (p2Flag == OUTSIDE)
          {
            this->EdgeRemove(edge, iDir);
          }
        }

        iEdge = edge->next[iDir];
        iDir = edge->dirNext[iDir];
      }

      // loop over edges in vertex again
      // iEdge = this edge, jEdge = next edge
      // p1 = last pt in iEdge, pt = first pt in jEdge
      // if p1 != p2, add edge between them
      this->Edges.reserve(this->nEdges + vert.nEdge);
      iEdge = vert.first;
      iDir = vert.dirFirst;

      for (int i = 0; i < vert.nEdge; ++i)
      {
        edge = &this->Edges[iEdge];
        jEdge = edge->next[iDir];
        jDir = edge->dirNext[iDir];
        if (jEdge < 0)
        {
          jEdge = vert.first;
          jDir = vert.dirFirst;
        }

        p1 = (iDir == 0) ? edge->p2.data() : edge->p1.data();
        p2 = (jDir == 0) ? this->Edges[jEdge].p1.data() : this->Edges[jEdge].p2.data();

        if (!this->SamePoint(p1, p2, this->Tolerance))
        {
          this->Edges.push_back(Edge());
          newEdge = &this->Edges.back();
          int n = this->nEdges++;
          memcpy(newEdge->p1.data(), p1, 3 * sizeof(double));
          memcpy(newEdge->p2.data(), p2, 3 * sizeof(double));

          // convert jEdge back to -1 for last vertex
          if (jEdge == vert.first)
          {
            jEdge = -1;
          }
          this->EdgeInsert(n, 0, iVert, iEdge, iDir, jEdge, jDir);
          ++i;
        }

        iEdge = jEdge;
        iDir = jDir;
      }
      ++iVert;
    }
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::ClipAdjust(vtkIdType cellId, double* lo, double* hi)
{
  double pBoth[3], move1[3], move2[3];
  int faces1[6], faces2[6];
  double* p1, * p2;

  // epsilon = EpsilonEdge fraction of largest cell dimension
  double epsilon = this->EpsilonEdge * (hi[0] - lo[0]);
  epsilon = std::max(epsilon, this->EpsilonEdge * (hi[1] - lo[1]));
  epsilon = std::max(epsilon, this->EpsilonEdge * (hi[2] - lo[2]));

  // collapse edges shorter than epsilon to a single point (so will be removed)
  // one or both of the points should be on cell faces
  // tiny edges can occur for several reasons:
  // (1) pre-clip end point is within epsilon of face
  // (2) a shell barely grazes cell edge or corner
  // (3) clipping round-off produces tiny edges near cell faces
  vtkIdType iEdge = 0;
  for (auto& edge : this->Edges)
  {
    if (!edge.active)
    {
      ++iEdge;
      continue;
    }

    p1 = edge.p1.data();
    p2 = edge.p2.data();
    double dxyz[3] = { 0.0, 0.0, 0.0 };
    vtkMath::Subtract(p1, p2, dxyz);
    double edgeLen = std::sqrt(vtkMath::SquaredNorm(dxyz));

    int nTiny = 0;
    if (edgeLen < epsilon)
    {
      ++nTiny;

      int nFace1 = this->OnFaces(p1, lo, hi, faces1);
      int nFace2 = this->OnFaces(p2, lo, hi, faces2);

      // set both p1 and p2 to same pboth
      // if both pts are interior (should not happen), pboth = p1
      // if only one pt X is on a face, pboth = X
      // if both pts are on one or more faces:
      // push both to face(s), recalculate OnFaces()
      // if pt X is on more faces, pboth = X, else pboth = p1
      if (!nFace1 && !nFace2)
      {
        memcpy(pBoth, p1, 3 * sizeof(double));
      }
      else if (nFace1 && !nFace2)
      {
        memcpy(pBoth, p1, 3 * sizeof(double));
      }
      else if (nFace2 && !nFace1)
      {
        memcpy(pBoth, p2, 3 * sizeof(double));
      }
      else
      {
        memcpy(move1, p1, 3 * sizeof(double));
        memcpy(move2, p2, 3 * sizeof(double));
        this->MoveToFaces(move1, lo, hi, epsilon);
        this->MoveToFaces(move2, lo, hi, epsilon);
        nFace1 = this->OnFaces(move1, lo, hi, faces1);
        nFace2 = this->OnFaces(move2, lo, hi, faces2);

        if (nFace2 > nFace1)
        {
          memcpy(pBoth, move2, 3 * sizeof(double));
        }
        else
        {
          memcpy(pBoth, move1, 3 * sizeof(double));
        }
      }

      // Set all points that are same as old p1 or p2 to pBoth
      // reset first for all jEdge != iEdge, then reset iEdge
      for (vtkIdType jEdge = 0; jEdge < this->nEdges; ++jEdge)
      {
        auto& edgeJ = this->Edges[jEdge];
        if (!edgeJ.active)
        {
          continue;
        }
        if (jEdge == iEdge)
        {
          continue;
        }

        if (this->SamePoint(edgeJ.p1.data(), p1, this->Tolerance))
        {
          memcpy(edgeJ.p1.data(), pBoth, 3 * sizeof(double));
        }
        if (this->SamePoint(edgeJ.p2.data(), p1, this->Tolerance))
        {
          memcpy(edgeJ.p2.data(), pBoth, 3 * sizeof(double));
        }

        if (this->SamePoint(edgeJ.p1.data(), p2, this->Tolerance))
        {
          memcpy(edgeJ.p1.data(), pBoth, 3 * sizeof(double));
        }
        if (this->SamePoint(edgeJ.p2.data(), p2, this->Tolerance))
        {
          memcpy(edgeJ.p2.data(), pBoth, 3 * sizeof(double));
        }
      }

      memcpy(edge.p1.data(), pBoth, 3 * sizeof(double));
      memcpy(edge.p2.data(), pBoth, 3 * sizeof(double));
    }
    ++iEdge;
  }

  // remove zero-length edges
  for (auto& edge : this->Edges)
  {
    if (!edge.active)
    {
      continue;
    }
    if (this->SamePoint(edge.p1.data(), edge.p2.data(), this->Tolerance))
    {
      this->EdgeRemove(&edge);
    }
  }

  // remove vertices (shells) which now have less than 3 edges
  // do this after deleting zero-length edges so vertices are updated
  // removals should have 2 or 0 edges, no verts should have 1 edge
  // if all shells are removed, BPG will be empty,
  // which will result in cell corner pts being left UNKNOWN in split().
  // to try and avoid this, tally the inside/outside for all removed shells
  // shell is "outside" if it implies cell is outside the surf (in the flow)
  // shell is "inside" if it implies cell is inside the surf (not in the flow)
  // some tris may not follow this rule, but most should
  // outside = shell norm from shell ctr points away from cell ctr
  // inside = shell norm from shell ctr points away from cell ctr 
  // centerOfBox = cell center pt
  // centerOfShell = shell center pt
  // shell2b = centerOfBox-centerOfShell
  vtkIdType nOutside = 0;
  vtkIdType nInside = 0;
  double centerOfBox[3], shell2b[3];

  vtkPoints* points = this->Source->GetPoints();
  vtkIdList* idList = vtkIdList::New();

  for (auto& vert : this->Verts)
  {
    if (vert.nEdge > 2)
    {
      continue;
    }

    vtkMath::Add(lo, hi, centerOfBox);
    vtkMath::MultiplyScalar(centerOfBox, 0.5);

    vtkIdType iShell = this->CellSurfs->GetId(vert.label);

    vtkCell* shell = this->Source->GetCell(iShell);
    idList = shell->GetPointIds();
    int idNum = idList->GetNumberOfIds();
    double centerOfShell[3] = {0.0, 0.0, 0.0};
    vtkPolygon::ComputeCentroid(points, idNum, idList->GetPointer(0), centerOfShell);
    vtkMath::Subtract(centerOfBox, centerOfShell, shell2b);
    double dot = vtkMath::Dot(vert.norm, shell2b);
    if (dot > 0.0)
    {
      ++nOutside;
    }
    if (dot < 0.0)
    {
      ++nInside;
    }

    this->VertexRemove(&vert);
  }

  // discard clipped shell if lies on a cell face w/ normal out of cell
  // increment GrazeCount in this case
  this->TouchCount = 0;
  this->GrazeCount = 0;
  int grazeFlag = 0;
  for (auto& vert : this->Verts)
  {
    if (!vert.active)
    {
      continue;
    }

    if (this->Grazing(&vert, lo, hi))
    {
      this->VertexRemove(&vert);
      this->GrazeCount++;
    }
  }

  // remove edges which now have no vertices
  for (auto& edge : this->Edges) {
    if (!edge.active)
    {
      continue;
    }
    if (edge.nVert == 0)
    {
      edge.active = 0;
    }
  }

  // set BPG empty flag if no active vertices
  this->Empty = 1;
  for (const auto& vert : this->Verts)
  {
    if (vert.active)
    {
      this->Empty = 0;
      break;
    }
  }

  // if no lines, set TouchMark which will be used to mark corner points
  // only set TouchMark if all deleted tris had same orientation
  // NOTE: setting TouchMark based on orientation of just majority
  // triggered later corner marking error for cone_test/in.cone problem 
  this->TouchMark = UNKNOWN;
  if (this->Empty)
  {
    if (nInside && nOutside == 0)
    {
      this->TouchMark = INSIDE;
    }
    else if (nOutside && nInside == 0)
    {
      this->TouchMark = OUTSIDE;
    }
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::CellShellVolume(double* lo, double* hi)
{
  vtkIdType iEdge, nEdge;
  int iDir;
  double zArea, volume;
  double* p0, * p1, * p2;
  Edge* edge;

  for (auto& vert : this->Verts)
  {
    if (!vert.active)
    {
      continue;
    }

    iEdge = vert.first;
    iDir = vert.dirFirst;
    nEdge = vert.nEdge;

    p0 = (iDir == 0) ? this->Edges[iEdge].p1.data() : this->Edges[iEdge].p2.data();

    volume = 0.0;
    for (vtkIdType i = 0; i < nEdge; ++i)
    {
      edge = &this->Edges[iEdge];

      // compute projected volume of a convex polygon to zlo face
      // split polygon into triangles
      // each tri makes a tri-capped volume with zlo face
      // zArea = area of oriented tri projected into z plane
      // volume based on height of z midpt of tri above zlo face
      p1 = edge->p1.data();
      p2 = edge->p2.data();
      if (iDir != 0)
      {
        std::swap(p1, p2);
      }

      zArea = 0.5 * ((p1[0] - p0[0]) * (p2[1] - p0[1]) - (p1[1] - p0[1]) * (p2[0] - p0[0]));
      volume -= zArea * ((p0[2] + p1[2] + p2[2]) / 3.0 - lo[2]);

      iEdge = edge->next[iDir];
      iDir = edge->dirNext[iDir];
    }

    vert.volume = volume;
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::EdgeToFace(double* lo, double* hi)
{
  vtkIdType iVert;
  int iFace;
  int faces[6] = {0, 0, 0, 0, 0, 0};
  std::array<double, 3> normInward;

  // insure each facelist has sufficient length
  vtkIdType nEdge = this->nEdges;
  for (int i = 0; i < 6; ++i)
  {
    this->FaceList[i].clear();
    this->FaceList[i].reserve(nEdge);
  }

  // loop over edges, assign singlets to exactly one face
  vtkIdType iEdge = 0;
  for (auto& edge : this->Edges)
  {
    if (!edge.active)
    {
      ++iEdge;
      continue;
    }

    if (edge.nVert == 3)
    {
      ++iEdge;
      continue;
    }

    int nFace = this->WhichFaces(edge.p1.data(), edge.p2.data(), lo, hi, faces);
    if (nFace == 0)
    {
      return 2;
    }
    else if (nFace == 1)
    {
      iFace = faces[0];
    }
    else if (nFace == 2)
    {
      iFace = faces[0];
      normInward.fill(0.0);
      normInward[iFace / 2] = (iFace % 2) ? -1.0 : 1.0;
      iVert = (edge.nVert == 1) ? edge.verts[0] : edge.verts[1];
      double dot = vtkMath::Dot(normInward.data(), this->Verts[iVert].norm);
      if (dot > 0.0)
      {
        iFace = faces[1];
      }
    }
    else
    {
      return 3;
    }

    this->FaceList[iFace].emplace_back(iEdge++);
  }

  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::EdgeToCLines(int iFace)
{
  vtkIdType iEdge;
  double* p1, * p2;
  Edge* edge;
  Cline* cline;

  int flip = 0;
  if (iFace == 0 || iFace == 3 || iFace == 4)
  {
    flip = 1;
  }

  vtkIdType nLine = this->FaceList[iFace].size();
  this->CLines.clear();
  this->CLines.resize(nLine);

  for (vtkIdType i = 0; i < nLine; ++i)
  {
    iEdge = this->FaceList[iFace][i];
    edge = &this->Edges[iEdge];
    p1 = edge->p1.data();
    p2 = edge->p2.data();
    if (edge->nVert != 1)
    {
      std::swap(p1, p2);
    }

    cline = &(this->CLines[i]);
    cline->line = iEdge;
    this->Compress2D(iFace, p1, cline->p1.data());
    this->Compress2D(iFace, p2, cline->p2.data());
    if (flip)
    {
      std::swap(cline->p1, cline->p2);
    }
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::AddFacePolygons(int iFace, double* lo, double* hi)
{
  vtkIdType mLoop, nLoop, mPt, nPt;
  vtkIdType iEdge, prev;
  int dir, dirPrev;
  double p1[3], p2[3];
  Vertex* vert;
  Edge* edge;
  Loop2D* loop;
  Point* p12d, * p22d;

  int flip = 0;
  if (iFace == 0 || iFace == 3 || iFace == 4)
  {
    flip = 1;
  }

  double value;
  int dim = iFace / 2;
  int lohi = iFace % 2;
  value = (lohi == 0) ? lo[dim] : hi[dim];

  vtkIdType nPg = this->nPolygons;
  vtkIdType nVert = this->nVerts;
  this->Verts.resize(nVert + nPg);

  for (auto& pg : this->Polygons)
  {
    vert = &this->Verts[nVert];
    vert->active = 1;
    vert->style = FACEPGON;
    vert->label = iFace;
    vert->volume = (iFace == 5) ? pg.area * (hi[2] - lo[2]) : 0.0;
    vert->nEdge = 0;

    prev = -1;
    dirPrev = -1;

    nLoop = pg.n;
    mLoop = pg.first;
    for (vtkIdType iLoop = 0; iLoop < nLoop; ++iLoop)
    {
      loop = &this->Loop2d[mLoop];
      nPt = loop->n;
      mPt = loop->first;
      this->Edges.reserve(this->nEdges + nPt);

      for (vtkIdType iPt = 0; iPt < nPt; ++iPt)
      {
        p12d = &this->Points2d[mPt];
        mPt = p12d->next;
        p22d = &this->Points2d[mPt];
        this->Expand2D(iFace, value, p12d->x.data(), p1);
        this->Expand2D(iFace, value, p22d->x.data(), p2);

        // edge was from a CSHELL vertex
        // match in opposite order that CSHELL vertex matched it
        if (p12d->type == ENTRY || p12d->type == TWO)
        {
          iEdge = p12d->line;
          edge = &this->Edges[iEdge];
          edge->style = CSHELLFACE;
          dir = (edge->nVert == 1) ? 1 : 0;
          this->EdgeInsert(iEdge, dir, nVert, prev, dirPrev, -1, -1);
          prev = iEdge;
          dirPrev = dir;
          continue;
        }

        // face edge not from a CSHELL
        // unflip edge if in a flip face
        iEdge = (flip) ? this->FindEdge(p2, p1, 0, dir)
          : this->FindEdge(p1, p2, 0, dir);
        if (iEdge == -2)
        {
          return 1;
        }

        if (iEdge >= 0)
        {
          this->EdgeInsert(iEdge, dir, nVert, prev, dirPrev, -1, -1);
          prev = iEdge;
          dirPrev = 1;
          continue;
        }

        iEdge = this->nEdges++;
        this->Edges.push_back(Edge());
        edge = &this->Edges.back();
        edge->style = FACEPGON;
        edge->nVert = 0;
        memcpy(edge->p1.data(), p1, 3 * sizeof(double));
        memcpy(edge->p2.data(), p2, 3 * sizeof(double));
        if (flip)
        {
          std::swap(edge->p1, edge->p2);
        }
        dir = 0;
        this->EdgeInsert(iEdge, dir, nVert, prev, dirPrev, -1, -1);
        prev = iEdge;
        dirPrev = 0;
      }
      mLoop = loop->next;
    }

    ++nVert;
  }

  this->nVerts = nVert;

  return 0;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::AddFace(int iFace,
  double* lo, double* hi, double* lo2d, double* hi2d)
{
  vtkIdType iEdge, prev;
  int dir, dirPrev;
  double p1[3], p2[3];
  Vertex* vert;
  Edge* edge;

  vtkIdType nVert = this->nVerts++;
  this->Verts.reserve(nVert + 1);
  this->Verts.push_back(Vertex());
  vert = &this->Verts.back();
  vert->active = 1;
  vert->style = FACE;
  vert->label = iFace;
  vert->volume = (iFace == 5) ?
    (hi[0] - lo[0]) * (hi[1] - lo[1]) * (hi[2] - lo[2]) : 0.0;

  int dim = iFace / 2;
  int lohi = iFace % 2;
  double value = (lohi == 0) ? lo[dim] : hi[dim];

  // usual ordering of points in face as LL,LR,UR,UL
  // flip order if in a flip face
  int flip = (iFace == 0 || iFace == 3 || iFace == 4) ? 1 : 0;

  double cPts[4][2];

  if (flip)
  {
    cPts[0][0] = lo2d[0]; cPts[0][1] = lo2d[1];
    cPts[1][0] = lo2d[0]; cPts[1][1] = hi2d[1];
    cPts[2][0] = hi2d[0]; cPts[2][1] = hi2d[1];
    cPts[3][0] = hi2d[0]; cPts[3][1] = lo2d[1];
  }
  else
  {
    cPts[0][0] = lo2d[0]; cPts[0][1] = lo2d[1];
    cPts[1][0] = hi2d[0]; cPts[1][1] = lo2d[1];
    cPts[2][0] = hi2d[0]; cPts[2][1] = hi2d[1];
    cPts[3][0] = lo2d[0]; cPts[3][1] = hi2d[1];
  }

  if (vert->nEdge)
  {
    prev = vert->last;
    dirPrev = vert->dirLast;
  }
  else
  {
    prev = -1;
    dirPrev = -1;
  }

  this->Edges.reserve(this->nEdges + 4);

  for (int i = 0; i < 4; ++i)
  {
    int j = (i + 1) % 4;

    this->Expand2D(iFace, value, &cPts[i][0], p1);
    this->Expand2D(iFace, value, &cPts[j][0], p2);
    iEdge = this->FindEdge(p1, p2, 1, dir);
    if (iEdge == -2)
    {
      return 1;
    }

    if (iEdge >= 0)
    {
      this->EdgeInsert(iEdge, dir, nVert, prev, dirPrev, -1, -1);
      prev = iEdge;
      dirPrev = 1;
      continue;
    }

    iEdge = this->nEdges++;
    this->Edges.push_back(Edge());
    edge = &this->Edges.back();
    edge->style = vert->style;
    memcpy(edge->p1.data(), p1, 3 * sizeof(double));
    memcpy(edge->p2.data(), p2, 3 * sizeof(double));
    dir = 0;
    this->EdgeInsert(iEdge, dir, nVert, prev, dirPrev, -1, -1);
    prev = iEdge;
    dirPrev = 0;
  }

  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::RemoveFaces()
{
  vtkIdType iEdge;
  int i, dir;
  Vertex* vert;
  Edge* edge;
  vtkIdType nVert = this->nVerts;

  for (int iter = 0; iter < 2; ++iter)
  {
    for (vtkIdType iVert = 0; iVert < nVert; ++iVert)
    {
      if (!this->Verts[iVert].active)
      {
        continue;
      }
      if (this->Verts[iVert].style != FACE)
      {
        continue;
      }
      vert = &this->Verts[iVert];

      iEdge = vert->first;
      dir = vert->dirFirst;
      for (i = 0; i < 4; ++i)
      {
        edge = &this->Edges[iEdge];
        if (edge->nVert == 1 || edge->nVert == 2)
        {
          break;
        }
        iEdge = edge->next[dir];
        dir = edge->dirNext[dir];
      }
      if (i < 4)
      {
        this->VertexRemove(vert);
      }
    }
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::Check()
{
  vtkIdType iEdge, last;
  int dir, dirLast;
  Vertex* vert;
  Edge* edge;

  // mark all edges as unclipped
  // use for detecting duplicate edges in same vertex
  vtkIdType nEdge = this->nEdges;
  for (vtkIdType iEdge = 0; iEdge < nEdge; ++iEdge)
  {
    if (this->Edges[iEdge].active)
    {
      this->Edges[iEdge].clipped = 0;
    }
  }

  // check vertices
  // for each vertex: mark edges as see them, unmark all edges at end
  vtkIdType nVert = this->nVerts;
  for (vtkIdType iVert = 0; iVert < nVert; ++iVert)
  {
    if (!this->Verts[iVert].active)
    {
      continue;
    }
    vert = &this->Verts[iVert];
    if (vert->nEdge < 3)
    {
      return 11;
    }

    nEdge = vert->nEdge;
    iEdge = vert->first;
    dir = vert->dirFirst;

    for (vtkIdType i = 0; i < nEdge; ++i)
    {
      edge = &this->Edges[iEdge];
      if (!edge->active)
      {
        return 12;
      }
      if (edge->verts[dir] != iVert)
      {
        return 13;
      }
      if (edge->clipped)
      {
        return 14;
      }
      edge->clipped = 1;
      last = iEdge;
      dirLast = dir;
      iEdge = edge->next[dir];
      dir = edge->dirNext[dir];
    }

    if (last != vert->last || dirLast != vert->dirLast)
    {
      return 15;
    }

    iEdge = vert->first;
    dir = vert->dirFirst;
    for (vtkIdType i = 0; i < nEdge; ++i)
    {
      edge = &this->Edges[iEdge];
      edge->clipped = 0;
      iEdge = edge->next[dir];
      dir = edge->dirNext[dir];
    }
  }

  // check edges
  nEdge = this->nEdges;
  for (vtkIdType iEdge = 0; iEdge < nEdge; ++iEdge)
  {
    if (!this->Edges[iEdge].active)
    {
      continue;
    }
    if (edge->nVert != 3)
    {
      return 16;
    }
    if (edge->verts[0] == edge->verts[1])
    {
      return 17;
    }
    if (edge->verts[0] >= nVert || !this->Verts[edge->verts[0]].active)
    {
      return 18;
    }
    if (edge->verts[1] >= nVert || !this->Verts[edge->verts[1]].active)
    {
      return 19;
    }
  }

  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::Walk()
{
  vtkIdType iVert, firstVert, iEdge, nEdge;
  int dir;
  Vertex* vert;
  Edge* edge;

  // used = 0/1 flag for whether a vertex is already part of a loop
  // only active vertices are eligible
  vtkIdType nVert = this->nVerts;
  this->Used.clear();
  this->Used.resize(nVert, 1);
  for (vtkIdType iVert = 0; iVert < nVert; ++iVert)
  {
    if (this->Verts[iVert].active)
    {
      this->Used[iVert] = 0;
    }
  }

  // stack = list of vertex indices to process
  // max size = # of vertices
  this->Stack.clear();
  this->Stack.resize(nVert, -1);
  int nStack = 0;

  // iterate over all vertices
  // start a loop at any unused vertex
  // add more vertices to loop via stack
  // check all neighbor vertices via shared edges
  // if neighbor vertex is unused, add to stack
  // stop when stack is empty
  vtkIdType nLoop = 0;

  for (vtkIdType i = 0; i < nVert; ++i)
  {
    if (this->Used[i])
    {
      continue;
    }

    double volume = 0.0;
    int flag = INTERIOR;
    vtkIdType nCount = 0;

    this->Stack[0] = firstVert = i;
    nStack = 1;
    this->Used[i] = 1;
    vtkIdType prev = -1;

    while (nStack)
    {
      --nStack;
      iVert = this->Stack[nStack];
      ++nCount;

      vert = &this->Verts[iVert];
      if (vert->style != CSHELL)
      {
        flag = BORDER;
      }
      volume += vert->volume;

      nEdge = vert->nEdge;
      iEdge = vert->first;
      dir = vert->dirFirst;

      for (vtkIdType i = 0; i < nEdge; ++i)
      {
        edge = &this->Edges[iEdge];
        if (!this->Used[edge->verts[0]])
        {
          this->Stack[nStack++] = edge->verts[0];
          this->Used[edge->verts[0]] = 1;
        }
        if (!this->Used[edge->verts[1]])
        {
          this->Stack[nStack++] = edge->verts[1];
          this->Used[edge->verts[1]] = 1;
        }
        iEdge = edge->next[dir];
        dir = edge->dirNext[dir];
      }

      if (prev >= 0)
      {
        this->Verts[prev].next = iVert;
      }
      prev = iVert;
    }
    vert->next = -1;

    this->Loop3d.resize(nLoop + 1);
    this->Loop3d[nLoop].volume = volume;
    this->Loop3d[nLoop].flag = flag;
    this->Loop3d[nLoop].n = nCount;
    this->Loop3d[nLoop].first = firstVert;
    ++nLoop;
  }

  this->nLoop3d = nLoop;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::LoopToPolyhedron()
{
  vtkIdType positive = 0;
  vtkIdType negative = 0;

  vtkIdType nLoop = this->nLoop3d;
  for (auto& loop3d : this->Loop3d)
  {
    if (loop3d.volume > 0.0)
    {
      ++positive;
    }
    else
    {
      ++negative;
    }
  }

  if (positive == 0)
  {
    return 4;
  }
  if (positive > 1 && negative)
  {
    return 5;
  }

  this->Polyhedons.resize(positive);

  if (positive == 1)
  {
    double volume = 0.0;
    for (vtkIdType i = 0; i < nLoop; ++i)
    {
      volume += this->Loop3d[i].volume;
      this->Loop3d[i].next = i + 1;
    }
    this->Loop3d[nLoop3d - 1].next = -1;

    if (volume < 0.0)
    {
      return 6;
    }

    this->Polyhedons[0].volume = volume;
    this->Polyhedons[0].n = nLoop3d;
    this->Polyhedons[0].first = 0;
  }
  else
  {
    for (vtkIdType i = 0; i < nLoop; ++i)
    {
      this->Polyhedons[i].volume = this->Loop3d[i].volume;
      this->Polyhedons[i].n = 1;
      this->Polyhedons[i].first = i;
      this->Loop3d[i].next = -1;
    }
  }

  this->nPolyhedrons = positive;

  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::CreateSurfMap()
{
  this->SurfMap.clear();
  this->SurfMap.resize(this->nSurf, -1);

  vtkIdType nPH = this->nPolyhedrons;
  for (vtkIdType iPH = 0; iPH < nPH; ++iPH)
  {
    vtkIdType nLoop = this->Polyhedons[iPH].n;
    vtkIdType mLoop = this->Polyhedons[iPH].first;
    for (vtkIdType iLoop = 0; iLoop < nLoop; ++iLoop)
    {
      vtkIdType nVert = this->Loop3d[mLoop].n;
      vtkIdType mVert = this->Loop3d[mLoop].first;
      for (vtkIdType iVert = 0; iVert < nVert; ++iVert)
      {
        if (this->Verts[mVert].style == CSHELL || this->Verts[mVert].style == CSHELLFACE)
        {
          this->SurfMap[this->Verts[mVert].label] = iPH;
        }
        mVert = this->Verts[mVert].next;
      }
      mLoop = this->Loop3d[mLoop].next;
    }
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::EdgeInsert(vtkIdType iEdge, int dir, vtkIdType iVert,
  vtkIdType iPrev, int dirPrev, vtkIdType iNext, int dirNext)
{
  Edge& edge = this->Edges[iEdge];
  Vertex& vert = this->Verts[iVert];

  if (dir == 0)
  {
    edge.nVert += 1;
    edge.verts[0] = iVert;
  }
  else
  {
    edge.nVert += 2;
    edge.verts[1] = iVert;
  }

  edge.active = 1;
  edge.clipped = 0;

  // set prev/next pointers for doubly linked list of edges
  edge.next[dir] = iNext;
  edge.prev[dir] = iPrev;

  if (iNext >= 0)
  {
    edge.dirNext[dir] = dirNext;
    Edge* next = &this->Edges[iNext];
    next->prev[dirNext] = iEdge;
    next->dirPrev[dirNext] = dir;
  }
  else
  {
    edge.dirNext[dir] = -1;
  }

  if (iPrev >= 0)
  {
    edge.dirPrev[dir] = dirPrev;
    Edge* prev = &this->Edges[iPrev];
    prev->next[dirPrev] = iEdge;
    prev->dirNext[dirPrev] = dir;
  }
  else
  {
    edge.dirPrev[dir] = -1;
  }

  // add edge info to owing vertex
  vert.nEdge++;
  if (iPrev < 0)
  {
    vert.first = iEdge;
    vert.dirFirst = dir;
  }
  if (iNext < 0)
  {
    vert.last = iEdge;
    vert.dirLast = dir;
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::EdgeRemove(Edge* edge)
{
  if (edge->verts[0] >= 0)
  {
    this->EdgeRemove(edge, 0);
  }
  if (edge->verts[1] >= 0)
  {
    this->EdgeRemove(edge, 1);
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::EdgeRemove(Edge* edge, int dir)
{
  vtkIdType iVert = edge->verts[dir];
  edge->verts[dir] = -1;
  if (dir == 0)
  {
    --edge->nVert;
  }
  else
  {
    edge->nVert -= 2;
  }

  if (edge->nVert == 0)
  {
    edge->active = 0;
  }

  // reset prev/next pointers for doubly linked list to skip this edge
  if (edge->prev[dir] >= 0)
  {
    Edge* prev = &this->Edges[edge->prev[dir]];
    int dirPrev = edge->dirPrev[dir];
    prev->next[dirPrev] = edge->next[dir];
    prev->dirNext[dirPrev] = edge->dirNext[dir];
  }

  if (edge->next[dir] >= 0)
  {
    Edge* next = &this->Edges[edge->next[dir]];
    int dirNext = edge->dirNext[dir];
    next->prev[dirNext] = edge->prev[dir];
    next->dirPrev[dirNext] = edge->dirPrev[dir];
  }

  // update vertex for removal of this edge
  Vertex& vert = this->Verts[iVert];
  vert.nEdge--;
  if (edge->prev[dir] < 0)
  {
    vert.first = edge->next[dir];
    vert.dirFirst = edge->dirNext[dir];
  }
  if (edge->next[dir] < 0)
  {
    vert.last = edge->prev[dir];
    vert.dirLast = edge->dirPrev[dir];
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::VertexRemove(Vertex* vert)
{
  vert->active = 0;

  vtkIdType iEdge = vert->first;
  int dir = vert->dirFirst;
  vtkIdType nEdge = vert->nEdge;

  for (vtkIdType i = 0; i < nEdge; ++i)
  {
    Edge* edge = &this->Edges[iEdge];
    if (dir == 0)
    {
      edge->nVert--;
    }
    else
    {
      edge->nVert -= 2;
    }
    if (edge->nVert == 0)
    {
      edge->active = 0;
    }

    edge->verts[dir] = -1;
    iEdge = edge->next[dir];
    dir = edge->dirNext[dir];
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::Grazing(Vertex* vert, double* lo, double* hi)
{
  int count[6] = {0, 0, 0, 0, 0, 0};
  double* p = nullptr;

  vtkIdType iEdge = vert->first;
  int iDir = vert->dirFirst;
  vtkIdType nEdge = vert->nEdge;

  for (vtkIdType i = 0; i < nEdge; ++i)
  {
    Edge& edge = this->Edges[iEdge];
    p = (iDir == 0) ? edge.p1.data() : edge.p2.data();

    for (int d = 0; d < 3; ++d)
    {
      if (std::abs(p[d] - lo[d]) < this->Tolerance) ++count[2 * d];
      if (std::abs(p[d] - hi[d]) < this->Tolerance) ++count[2 * d + 1];
    }
    iEdge = edge.next[iDir];
    iDir = edge.dirNext[iDir];
  }

  for (int d = 0; d < 3; ++d)
  {
    if (count[2 * d] == nEdge && vert->norm[d] < 0.0)
    {
      return 1;
    }
    if (count[2 * d + 1] == nEdge && vert->norm[0] > 0.0)
    {
      return 1;
    }
  }

  return 0;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::OnFaces(double* p, double* lo, double* hi, int* faces)
{
  int n = 0;
  for (int d = 0; d < 3; ++d)
  {
    if (std::abs(p[d] - lo[d]) < this->Tolerance)
    {
      faces[n++] = 2 * d;
    }
    if (std::abs(p[d] - hi[d]) < this->Tolerance)
    {
      faces[n++] = 2 * d + 1;
    }
  }

  return n;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::WhichFaces(double* p1, double* p2, double* lo, double* hi, int* faces)
{
  int n = 0;
  for (int d = 0; d < 3; ++d)
  {
    if (std::abs(p1[d] - lo[d]) < this->Tolerance && std::abs(p2[d] - lo[d]) < this->Tolerance)
    {
      faces[n++] = 2 * d;
    }
    if (std::abs(p1[d] - hi[d]) < this->Tolerance && std::abs(p2[d] - hi[d]) < this->Tolerance)
    {
      faces[n++] = 2 * d + 1;
    }
  }

  return n;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::FaceFromCell(int iFace, double* lo, double* hi,
  double* lo2d, double* hi2d)
{
  if (iFace < 2)
  {
    lo2d[0] = lo[1]; hi2d[0] = hi[1];
    lo2d[1] = lo[2]; hi2d[1] = hi[2];
  }
  else if (iFace < 4)
  {
    lo2d[0] = lo[0]; hi2d[0] = hi[0];
    lo2d[1] = lo[2]; hi2d[1] = hi[2];
  }
  else
  {
    lo2d[0] = lo[0]; hi2d[0] = hi[0];
    lo2d[1] = lo[1]; hi2d[1] = hi[1];
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::Compress2D(int iFace, double* p3, double* p2)
{
  if (iFace < 2)
  {
    p2[0] = p3[1];
    p2[1] = p3[2];
  }
  else if (iFace < 4)
  {
    p2[0] = p3[0];
    p2[1] = p3[2];
  }
  else
  {
    p2[0] = p3[0];
    p2[1] = p3[1];
  }
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::Expand2D(int iFace, double value, double* p2, double* p3)
{
  if (iFace < 2)
  {
    p3[0] = value;
    p3[1] = p2[0];
    p3[2] = p2[1];
  }
  else if (iFace < 4)
  {
    p3[0] = p2[0];
    p3[1] = value;
    p3[2] = p2[1];
  }
  else
  {
    p3[0] = p2[0];
    p3[1] = p2[1];
    p3[2] = value;
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::Clip(vtkIdType shellId, vtkPoints* points,
  const vtkBoundingBox& triBBox, const vtkBoundingBox& cellBBox)
{
  double** path;
  double** newPath;
  double point[3];
  int nPath;
  double value;
  double* s, * e;

  // If the triangle is entirely contained in the cell, nothing is done.
  if (triBBox.IsSubsetOf(cellBBox))
  {
    return 1;
  }

  double lo[3], hi[3];
  cellBBox.GetMinPoint(lo);
  cellBBox.GetMaxPoint(hi);

  auto ptIds = vtkSmartPointer<vtkIdList>::New();
  this->Source->GetCellPoints(shellId, ptIds);
  vtkIdType nNew = ptIds->GetNumberOfIds();

  // initial path = shell vertices
  for (vtkIdType i = 0; i < nNew; ++i)
  {
    points->GetPoint(ptIds->GetId(i), point);
    memcpy(this->Path1[i], point, 3 * sizeof(double));
  }

  // Clip a shell against each of 6 grid face planes
  for (int d = 0; d < 3; ++d)
  {
    path = this->Path1;
    newPath = this->Path2;
    nPath = nNew;
    nNew = 0;

    value = lo[d];
    s = path[nPath - 1];
    for (int i = 0; i < nPath; ++i)
    {
      e = path[i];
      if (e[d] >= value)
      {
        if (s[d] < value)
        {
          this->Between(s, e, d, value, newPath[nNew++]);
        }
        memcpy(newPath[nNew++], e, 3 * sizeof(double));
      }
      else if (s[d] >= value)
      {
        this->Between(e, s, d, value, newPath[nNew++]);
      }
      s = e;
    }
    if (!nNew)
    {
      return 0;
    }

    path = this->Path2;
    newPath = this->Path1;
    nPath = nNew;
    nNew = 0;

    value = hi[d];
    s = path[nPath - 1];
    for (int i = 0; i < nPath; ++i)
    {
      e = path[i];
      if (e[d] <= value)
      {
        if (s[d] > value)
        {
          this->Between(s, e, d, value, newPath[nNew++]);
        }
        memcpy(newPath[nNew++], e, 3 * sizeof(double));
      }
      else if (s[d] <= value)
      {
        this->Between(e, s, d, value, newPath[nNew++]);
      }
      s = e;
    }
    if (!nNew)
    {
      return 0;
    }
  }

  return nNew;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::Between(double* a, double* b, int dim, double value, double* c)
{
  if (dim == 0)
  {
    c[1] = a[1] + (value - a[dim]) / (b[dim] - a[dim]) * (b[1] - a[1]);
    c[2] = a[2] + (value - a[dim]) / (b[dim] - a[dim]) * (b[2] - a[2]);
    c[0] = value;
  }
  else if (dim == 1)
  {
    c[0] = a[0] + (value - a[dim]) / (b[dim] - a[dim]) * (b[0] - a[0]);
    c[2] = a[2] + (value - a[dim]) / (b[dim] - a[dim]) * (b[2] - a[2]);
    c[1] = value;
  }
  else
  {
    c[0] = a[0] + (value - a[dim]) / (b[dim] - a[dim]) * (b[0] - a[0]);
    c[1] = a[1] + (value - a[dim]) / (b[dim] - a[dim]) * (b[1] - a[1]);
    c[2] = value;
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::SamePoint(double* x, double* y, double t)
{
  return vtkMathUtilities::FuzzyCompare(x[0], y[0], t) &&
    vtkMathUtilities::FuzzyCompare(x[1], y[1], t) && vtkMathUtilities::FuzzyCompare(x[2], y[2], t);
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::Corner(double* pt, double* lo, double* hi)
{
  if (std::abs(pt[2] - lo[2]) < this->Tolerance)
  {
    if (std::abs(pt[1] - lo[1]) < this->Tolerance)
    {
      if (std::abs(pt[0] - hi[0]) < this->Tolerance)
      {
        return 0;
      }
      else if (std::abs(pt[0] - hi[0]) < this->Tolerance)
      {
        return 1;
      }
    }
    else if (std::abs(pt[1] - hi[1]) < this->Tolerance)
    {
      if (std::abs(pt[0] - lo[0]) < this->Tolerance)
      {
        return 2;
      }
      else if (std::abs(pt[0] - hi[0]) < this->Tolerance)
      {
        return 3;
      }
    }
  }
  else if (std::abs(pt[2] - hi[2]) < this->Tolerance)
  {
    if (std::abs(pt[1] - lo[1]) < this->Tolerance)
    {
      if (std::abs(pt[0] - lo[0]) < this->Tolerance)
      {
        return 4;
      }
      else if (std::abs(pt[0] - hi[0]) < this->Tolerance)
      {
        return 5;
      }
    }
    else if (std::abs(pt[1] - hi[1]) < this->Tolerance)
    {
      if (std::abs(pt[0] - lo[0]) < this->Tolerance)
      {
        return 6;
      }
      else if (std::abs(pt[0] - hi[0]) < this->Tolerance)
      {
        return 7;
      }
    }
  }

  return -1;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::MoveToFaces(double* pt, double* lo, double* hi, double epsilon)
{
  for (int i = 0; i < 3; ++i)
  {
    if (std::abs(pt[i] - lo[i]) < epsilon)
    {
      pt[i] = lo[i];
    }
    if (std::abs(pt[i] - hi[i]) < epsilon)
    {
      pt[i] = hi[i];
    }
  }
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::FindEdge(double* x, double* y, int flag, int& dir)
{
  double* p1, * p2;

  vtkIdType nEdge = this->nEdges;
  for (vtkIdType i = 0; i < this->nEdges; ++i)
  {
    auto& edge = this->Edges[i];
    if (!edge.active)
    {
      continue;
    }
    if (flag && (edge.style == CSHELL || edge.style == CSHELLFACE))
    {
      continue;
    }

    p1 = edge.p1.data();
    p2 = edge.p2.data();
    if (this->SamePoint(x, p1, this->Tolerance) && this->SamePoint(y, p2, this->Tolerance))
    {
      if (edge.nVert % 2 == 1)
      {
        return -2;
      }
      dir = 0;
      return i;
    }
    if (this->SamePoint(x, p2, this->Tolerance) && this->SamePoint(y, p1, this->Tolerance))
    {
      if (edge.nVert / 2 == 1)
      {
        return -2;
      }
      dir = 1;
      return i;
    }
  }

  dir = -1;
  return -1;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::SplitFace(int id, int iFace, double* lo, double* hi)
{
  int errFlag = this->WeilerBuild(lo, hi);
  if (errFlag)
  {
    return errFlag + 20;
  }

  this->WeilerLoops(lo, hi);
  errFlag = this->LoopToPG();
  if (errFlag)
  {
    return errFlag + 20;
  }

  return 0;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::WeilerBuild(double* lo, double* hi)
{
  vtkIdType j;
  vtkIdType firstPt, lastPt, nextPt;
  double* pt;

  // insure space in points for 2 pts per cline + 4 corner pts
  vtkIdType nLines = this->CLines.size();
  this->Points2d.clear();
  this->Points2d.resize(2 * nLines + 4);
  this->nPoint2d = 0;

  // add each CLine end pt to points
  // set x,type,next,line for each pt
  vtkIdType npt = 0;
  for (auto& cLine : CLines)
  {
    // 1st point in Cline
    pt = cLine.p1.data();
    for (j = 0; j < npt; ++j)
    {
      if (this->SamePoint(pt, this->Points2d[j].x.data(), this->Tolerance))
      {
        break;
      }
    }

    // pt already exists
    if (j < npt)
    {
      if (this->Points2d[j].type == ENTRY || this->Points2d[j].type == TWO)
      {
        return 1;
      }
      this->Points2d[j].type = TWO;
      this->Points2d[j].line = cLine.line;
      firstPt = j;
    }
    else // new pt
    {
      memcpy(this->Points2d[npt].x.data(), pt, 3 * sizeof(double));
      this->Points2d[npt].type = ENTRY;
      this->Points2d[npt].line = cLine.line;
      firstPt = npt++;
    }

    // 2nd point in cline
    pt = cLine.p2.data();
    for (j = 0; j < npt; ++j)
    {
      if (this->SamePoint(pt, this->Points2d[j].x.data(), this->Tolerance))
      {
        break;
      }
    }

    // pt already exists
    if (j < npt)
    {
      if (this->Points2d[j].type == EXIT || this->Points2d[j].type == TWO)
      {
        return 2;
      }
      this->Points2d[j].type = TWO;
      this->Points2d[firstPt].next = j;
    }
    else // new pt
    {
      memcpy(this->Points2d[npt].x.data(), pt, 3 * sizeof(double));
      this->Points2d[npt].type = EXIT;
      this->Points2d[firstPt].next = npt;
      ++npt;
    }
  }

  // error check that every singlet point is on cell border
  for (vtkIdType i = 0; i < npt; ++i)
  {
    if (this->Points2d[i].type != TWO &&
      this->PtFlag(this->Points2d[i].x.data(), lo, hi) != BORDER)
    {
      return 3;
    }
  }

  // add 4 cell CORNER pts to points
  // only if corner pt is not already an ENTRY or EXIT pt
  // if a TWO pt, still add corner pt as CORNER
  // ech corner pt is at beginning of side it is on
  // NOTE: could use hash to find existing pts in O(1) time
  // corner flag = 0,1,2,3 for LL,LR,UL,UR, same as in Grid::ChildInfo,
  //   but ordering of corner pts in linked list is LL,LR,UR,UL
  // side = 0,1,2,3 for lower,right,upper,left = traversal order in linked list
  // value = x-coord for lower/upper sides, y-coord for left,right sides
  double cPt[3] = {lo[0], lo[1], 0.0};
  vtkIdType ipt1, ipt2, ipt3, ipt4;

  for (j = 0; j < npt; ++j)
  {
    if (this->SamePoint(cPt, this->Points2d[j].x.data(), this->Tolerance))
    {
      break;
    }
  }
  if (j == npt || this->Points2d[j].type == TWO)
  {
    memcpy(this->Points2d[npt].x.data(), cPt, 3 * sizeof(double));
    this->Points2d[npt].type = CORNER;
    ipt1 = npt++;
  }
  else
  {
    ipt1 = j;
  }
  this->Points2d[ipt1].corner = 0;
  this->Points2d[ipt1].side = 0;
  this->Points2d[ipt1].value = lo[0];

  cPt[0] = hi[0]; cPt[1] = lo[1];
  for (j = 0; j < npt; ++j)
  {
    if (this->SamePoint(cPt, this->Points2d[j].x.data(), this->Tolerance))
    {
      break;
    }
  }
  if (j == npt || this->Points2d[j].type == TWO)
  {
    memcpy(this->Points2d[npt].x.data(), cPt, 3 * sizeof(double));
    this->Points2d[npt].type = CORNER;
    ipt2 = npt++;
  }
  else
  {
    ipt2 = j;
  }
  this->Points2d[ipt2].corner = 1;
  this->Points2d[ipt2].side = 1;
  this->Points2d[ipt2].value = lo[1];

  cPt[0] = hi[0];
  cPt[1] = hi[1];
  for (j = 0; j < npt; ++j)
  {
    if (this->SamePoint(cPt, this->Points2d[j].x.data(), this->Tolerance))
    {
      break;
    }
  }
  if (j == npt || this->Points2d[j].type == TWO)
  {
    memcpy(this->Points2d[npt].x.data(), cPt, 3 * sizeof(double));
    this->Points2d[npt].type = CORNER;
    ipt3 = npt++;
  }
  else
  {
    ipt3 = j;
  }
  this->Points2d[ipt3].corner = 3;
  this->Points2d[ipt3].side = 2;
  this->Points2d[ipt3].value = hi[0];

  cPt[0] = lo[0];
  cPt[1] = hi[1];
  for (j = 0; j < npt; ++j)
  {
    if (this->SamePoint(cPt, this->Points2d[j].x.data(), this->Tolerance))
    {
      break;
    }
  }
  if (j == npt || this->Points2d[j].type == TWO)
  {
    memcpy(this->Points2d[npt].x.data(), cPt, 3 * sizeof(double));
    this->Points2d[npt].type = CORNER;
    ipt4 = npt++;
  }
  else
  {
    ipt4 = j;
  }
  this->Points2d[ipt4].corner = 2;
  this->Points2d[ipt4].side = 3;
  this->Points2d[ipt4].value = hi[1];

  // n = final # of points
  this->nPoint2d = npt;

  // create initial counter-clockwise linked list around cell perimeter
  // just the 4 corner pts
  firstPt = ipt1;
  lastPt = ipt4;

  this->Points2d[ipt1].cprev = -1;
  this->Points2d[ipt1].cnext = ipt2;
  this->Points2d[ipt2].cprev = ipt1;
  this->Points2d[ipt2].cnext = ipt3;
  this->Points2d[ipt3].cprev = ipt2;
  this->Points2d[ipt3].cnext = ipt4;
  this->Points2d[ipt4].cprev = ipt3;
  this->Points2d[ipt4].cnext = -1;

  // add all non-corner ENTRY/EXIT pts to counter-clockwise linked list
  // side = 0,1,2,3 for lower,right,upper,left sides
  // value = coord of pt along the side it is on
  vtkIdType ipt, iPrev;
  int side;
  double value;

  for (vtkIdType i = 0; i < npt; ++i)
  {
    if (this->Points2d[i].type == TWO || this->Points2d[i].type == CORNER)
    {
      continue;
    }
    if (this->Points2d[i].corner >= 0)
    {
      continue;
    }

    side = this->WhichSide(this->Points2d[i].x.data(), lo, hi);
    value = (side % 2) ? this->Points2d[i].x[1] : this->Points2d[i].x[0];

    // interleave Ith point into linked list between firstPt and lastPt
    // insertion location is between iprev and ipt
    // special logic if inserting at end of list
    ipt = firstPt;
    while (ipt >= 0)
    {
      if (side < this->Points2d[ipt].side)
      {
        break;
      }
      if (side == this->Points2d[ipt].side)
      {
        if (side < 2)
        {
          if (value < this->Points2d[ipt].value)
          {
            break;
          }
        }
        else
        {
          if (value > this->Points2d[ipt].value)
          {
            break;
          }
        }
      }

      iPrev = ipt;
      ipt = this->Points2d[ipt].cnext;
    }

    this->Points2d[i].side = side;
    this->Points2d[i].value = value;
    this->Points2d[i].cprev = iPrev;
    this->Points2d[i].cnext = ipt;

    this->Points2d[iPrev].cnext = i;
    if (ipt >= 0)
    {
      this->Points2d[ipt].cprev = i;
    }
    else
    {
      lastPt = i;
    }
  }

  // set next field for cell perimeter points in linked list
  // this completes loops for next fields already set between ENTRY/EXIT pts
  // do not reset next for ENTRY pts
  // after loop, explicitly connect lastpt to firstpt
  ipt = firstPt;
  while (ipt >= 0)
  {
    nextPt = this->Points2d[ipt].cnext;
    if (this->Points2d[ipt].type != ENTRY)
    {
      this->Points2d[ipt].next = nextPt;
    }
    ipt = nextPt;
  }
  if (this->Points2d[lastPt].type != ENTRY)
  {
    this->Points2d[lastPt].next = firstPt;
  }

  // successful return
  return 0;
}

//-----------------------------------------------------------------------------
void vtkCutCell3D::WeilerLoops(double* lo, double* hi)
{
  // used = 0/1 flag for whether a point is already part of a loop
  vtkIdType n = this->nPoint2d;
  this->Used.clear();
  this->Used.resize(n, 0);

  // iterate over all pts
  // start a loop at any unused pt
  // walk loop via next field:
  //  mark pts as used
  //  compute area along the way
  // stop when reach initial pt:
  //  closed loop, add to loops data structure
  //  loop of just 4 corner pts is a valid loop
  // stop when reach used pt:
  //  discard loop, just traversed non-loop corner pts
  vtkIdType iPt, nCount, firstPt, nextPt;
  int iFlag, cFlag;
  double area;
  double *p1, *p2;

  vtkIdType nLoop = 0;
  for (vtkIdType i = 0; i < n; ++i)
  {
    if (this->Used[i])
    {
      continue;
    }
    area = 0.0;
    iFlag = cFlag = 1;
    nCount = 0;

    iPt = firstPt = i;
    p1 = this->Points2d[iPt].x.data();

    while (!this->Used[iPt])
    {
      this->Used[iPt] = 1;
      ++nCount;
      if (this->Points2d[iPt].type != TWO)
      {
        iFlag = 0;
      }
      if (this->Points2d[iPt].type != CORNER)
      {
        cFlag = 0;
      }
      nextPt = this->Points2d[iPt].next;
      p2 = this->Points2d[nextPt].x.data();
      area -= (0.5 * (p1[1] + p2[1]) - lo[1]) * (p2[0] - p1[0]);
      p1 = p2;
      iPt = nextPt;
      if (iPt == firstPt)
      {
        break;
      }
    }
    if (iPt != firstPt)
    {
      continue;
    }

    this->Loop2d.resize(nLoop + 1);
    Loop2D& loop2d = this->Loop2d.back();
    loop2d.area = area;
    loop2d.active = 1;
    if (iFlag)
    {
      loop2d.flag = INTERIOR;
    }
    else if (cFlag)
    {
      loop2d.flag = BORDER;
    }
    else
    {
      loop2d.flag = INTBORD;
    }
    loop2d.n = nCount;
    loop2d.first = firstPt;
    ++nLoop;
  }

  this->nLoop2d = nLoop;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::LoopToPG()
{
  vtkIdType positive = 0;
  vtkIdType negative = 0;

  vtkIdType nLoop = this->nLoop2d;
  for (auto& loop2d : this->Loop2d)
  {
    if (loop2d.area > 0.0)
    {
      ++positive;
    }
    else
    {
      ++negative;
    }
  }
  if (positive == 0)
  {
    return 4;
  }
  if (positive > 1 && negative)
  {
    return 5;
  }

  // if multiple positive loops, mark BORDER loop as inactive if exists
  // don't want entire cell border to be a loop in this case
  if (positive > 1)
  {
    for (auto& loop2d : this->Loop2d)
    {
      if (loop2d.flag == BORDER)
      {
        loop2d.active = 0;
        --positive;
      }
    }
  }

  this->Polygons.clear();
  this->Polygons.resize(positive);

  // positive = 1 means 1 Polygon with area = sum of all pos/neg loops
  // positive > 1 means each loop is a Polygon
  if (positive == 1)
  {
    double area = 0.0;
    vtkIdType prev = -1;
    vtkIdType count = 0;
    vtkIdType first = -1;

    for (vtkIdType i = 0; i < nLoop; ++i)
    {
      if (!this->Loop2d[i].active)
      {
        continue;
      }
      area += this->Loop2d[i].area;
      ++count;
      if (prev < 0)
      {
        first = i;
      }
      else
      {
        this->Loop2d[prev].next = i;
      }
      prev = i;
    }
    this->Loop2d[prev].next = -1;

    this->Polygons[0].area = area;
    this->Polygons[0].n = count;
    this->Polygons[0].first = first;
  }
  else
  {
    vtkIdType m = 0;
    for (vtkIdType i = 0; i < nLoop; ++i)
    {
      if (!this->Loop2d[i].active)
      {
        continue;
      }
      this->Polygons[m].area = this->Loop2d[i].area;
      this->Polygons[m].n = 1;
      this->Polygons[m].first = i;
      ++m;
      this->Loop2d[i].next = -1;
    }
  }

  this->nPolygons = positive;

  return 0;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::PtFlag(double* pt, double* lo, double* hi)
{
  double x = pt[0];
  double y = pt[1];
  if (lo[0] - x >= this->Tolerance || x - hi[0] >= this->Tolerance ||
    lo[1] - y >= this->Tolerance || y - hi[1] >= this->Tolerance)
  {
    return EXTERIOR;
  }
  if (x - lo[0] >= this->Tolerance && hi[0] - x >= this->Tolerance &&
    y - lo[1] >= this->Tolerance && hi[1] - y >= this->Tolerance)
  {
    return INTERIOR;
  }

  return BORDER;
}

//-----------------------------------------------------------------------------
int vtkCutCell3D::WhichSide(double* pt, double* lo, double* hi)
{
  if (std::abs(pt[0] - lo[0]) < this->Tolerance)
  {
    return 3;
  }
  if (std::abs(pt[0] - hi[0]) < this->Tolerance)
  {
    return 1;
  }
  if (std::abs(pt[1] - lo[1]) < this->Tolerance)
  {
    return 0;
  }
  if (std::abs(pt[1] - hi[1]) < this->Tolerance)
  {
    return 2;
  }

  return -1;
}
VTK_ABI_NAMESPACE_END
