// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
/**
 * @class   vtkCutCell3D
 * @brief   generate a 3D unstructured grid from PolyData using cut-cell method
 *
 * vtkCutCell3D is a filter that creates 3D unstructured grid from
 * watertight polydata using the cut-cell method. The normal vectors of the
 * shell elements comprising this polydata must be consistent. The choice of
 * whether the grid bounds left in the cut-cell want to be inside or outside the
 * polygon can be set by InsideOut. When UseInputBounds is turned off, 
 * GridBounds can be used to specify the grid boundaries for cut-cells.
 * The resolution of the grid bounds in x-, y-, and z-directions for performing 
 * cut-cell is set in GridDimensions vector (default value is 10 x 10 x 10).
 * When ComputeVolume is enabled (default is disabled), the volume of each
 * element created by the cut-cell is calculated and stored in an array with the
 * name specified in VolumeArrayName.
 *  
 * @warning
 * If the cut cell fails, setting the Shrink value (default value is 0) to a
 * small value, such as 1e-8, may avoid the failure.
 */

#ifndef vtkCutCell3D_h
#define vtkCutCell3D_h

//#include "vtkFiltersCoreModule.h" // For export macro
#include "vtkUnstructuredGridAlgorithm.h"

#include <array>
#include <unordered_map>
#include <vector>

#define VTK_ABI_NAMESPACE_BEGIN
#define VTK_ABI_NAMESPACE_END

class CutCellInfo;
class Edge;
class Vertex;
class Cline;
class Point;
class Loop2D;
class Loop3D;
class Polygon;
class Polyhedron;

VTK_ABI_NAMESPACE_BEGIN
class vtkBoundingBox;
class vtkCellArray;
class vtkCellLocator;
class vtkDataSetAttributes;
class vtkDoubleArray;
class vtkIdList;
class vtkImplicitPolyDataDistance;
class vtkIncrementalPointLocator;
class vtkPoints;
class vtkPolyData;
class vtkUnstructuredGrid;

//class VTKFILTERSCORE_EXPORT  vtkCutCell3D : public vtkUnstructuredGridAlgorithm
class vtkCutCell3D : public vtkUnstructuredGridAlgorithm
{
public:
  static vtkCutCell3D* New();
  vtkTypeMacro(vtkCutCell3D, vtkUnstructuredGridAlgorithm);
  void PrintSelf(ostream&, vtkIndent) override;

  ///@{
  /**
   * Set/Get whether the grid bounds used for cut cell is obtained from
   * Input bounds or not. By default it is set to true.
   */
  vtkSetMacro(UseInputBounds, vtkTypeBool);
  vtkGetMacro(UseInputBounds, vtkTypeBool);
  vtkBooleanMacro(UseInputBounds, vtkTypeBool);
  ///@}

  ///@{
  /**
   * Set/Get cut cell grid bounds. If (UseInputBounds == 1) then the grid
   * bounds won't be used.
   */
  vtkSetVector6Macro(GridBounds, double);
  vtkGetVector6Macro(GridBounds, double);
  ///@}

  ///@{
  /**
   * Set/Get cut cell grid dimension along each axis. Default will be [10,10,10]
   */
  vtkSetVector3Macro(GridDimensions, int);
  vtkGetVector3Macro(GridDimensions, int);
  ///@}

  ///@{
  /**
   * Set / Get a spatical locator for merging points. By default,
   * an instance of vtkMergePoints is used.
   */
  void SetLocator(vtkIncrementalPointLocator* locator);
  vtkGetObjectMacro(Locator, vtkIncrementalPointLocator);
  ///@}

  /**
   * Create default locator. Used to create one when none is specified. The
   * locator is used to merge coincident points.
   */
  void CreateDefaultLocator();

  ///@{
  /**
   * Get/Set the shrink fraction to use to find same points when adding points in
   * the cut cell process. Default is 1e-14.
   *
   * This is simply passed on to the internal vtkLocator used to merge points.
   * @sa `vtkLocator::SetTolerance`.
   */
  vtkSetClampMacro(Tolerance, double, 0.0, VTK_DOUBLE_MAX);
  vtkGetMacro(Tolerance, double);
  ///@}

  ///@{
  /**
   * Get/Set the shrink fraction for the specified grid cell. If the splitting fails,
   * this gets rid of pesky errors due to shell pts/edges on cell. Default is 0.0.
   */
  vtkSetClampMacro(Shrink, double, 0.0, VTK_DOUBLE_MAX);
  vtkGetMacro(Shrink, double);
  ///@}

  ///@{
  /**
   * Specify whether or not to compute volumes.
   * This option is enabled by default.
   */
  vtkSetMacro(ComputeVolume, vtkTypeBool);
  vtkGetMacro(ComputeVolume, vtkTypeBool);
  vtkBooleanMacro(ComputeVolume, vtkTypeBool);
  ///@}

  ///@{
  /**
   * Set/Get the name of the computed volume array.
   * Default name is Volume.
   */
  vtkSetStringMacro(VolumeArrayName);
  vtkGetStringMacro(VolumeArrayName);
  ///@}

  ///@{
  /**
   * Set/Get the InsideOut flag. When off, Implicit Function leaves a negative region.
   * Conversely, when InsidOut is On, Implicit Function leaves a positive region.
   * InsideOut is off by default.
   */
  vtkSetMacro(InsideOut, vtkTypeBool);
  vtkGetMacro(InsideOut, vtkTypeBool);
  vtkBooleanMacro(InsideOut, vtkTypeBool);
  ///@}

protected:
  vtkCutCell3D();
  ~vtkCutCell3D() override;

  int RequestData(vtkInformation*, vtkInformationVector**, vtkInformationVector*) override;

  int FillInputPortInformation(int, vtkInformation*) override;

  /**
   * Helper methods to generate a 3D cut cell
   */
  void AddHex(vtkIdType cellId, double* origin, double* size);
  void AddPolyhedron(vtkIdType cellId, double* lo, double* hi, vtkUnstructuredGrid* output);
  void CutCell(vtkIdType cellId, double* origin, double* size);

  /**
   * Compute intersections of a grid cell with all surfs
   * CellSurfs = indices into global surf list
   * return nSurf = # of surfs
   */
  int SurfToGrid(vtkIdType id, const vtkBoundingBox& cellBBox);

  /**
   * Calculate cut volume of a grid cell that contains nSurf tris
   * also calculate if cell is split into distinct sub-volumes by tris
   */
  int Split(vtkIdType cellId, double* lo, double* hi, int& xSub, double* xSplit);

  /**
   * Add each shell as vertex and edges to BPG
   * Add full edge even if outside cell, clipping comes later
   */
  int AddShells(vtkIdType cellId);

  /**
   * Clip collectoin of shell that overlap cell by 6 faces of cell
   * loop over faces, loop over shell, loop over edges in shell
   * edges fully outside the cell are removed
   * shared edges that intersect the cell are clipped consistently
   */
  void ClipShells(vtkIdType cellId, double* lo, double* hi);

  /**
   * Adjust the collection of clipped shells (vertices)
   * discard if clipped shell is a single point, increment TouchCount
   * TouchMark = corner point marking inferred from touching shell orientations
   * discard if grazes cell with outward normal, increment GrazeCount
   * if all clipped shells are discarded
   * set and return empty, TouchCount, GrazeCount, TouchMark
   */
  void ClipAdjust(vtkIdType cellId, double* lo, double* hi);

  /**
   * Compute volume of vertices
   * when called, only clippled shells exist
   */
  void CellShellVolume(double* lo, double* hi);

  /**
   * Assign all singlet edges to faces (0-5)
   * snglet edge must be on one or two faces, two if on cell edge
   * if along cell edge, assign to one of two faces based on
   * dot product of inward face norm and norm of shell containing edge
   */
  int EdgeToFace(double* lo, double* hi);

  /**
   * Build a 2d CLINES data structure
   * from all singlet edge assigned to iface (0-5)
   * order pts in edge for shell traversing edge in forward order
   * flip edge if in a flip face = faces 0,3,4
   * edge label in clines = edge index in BPG
   */
  void EdgeToCLines(int iFace);

  /**
   * Add one or more face polygons as vertices to BPG
   * have to convert pts computed by 2d back into 3d pts on face
   */
  int AddFacePolygons(int iFace, double* lo, double* hi);

  /**
   * Add an entire cell face as vertex to BPG
   * if outerflag2d = 0, create new veretx
   * else face polygon already exists, so add edges to it
   */
  int AddFace(int iFace, double* lo, double* hi, double* lo2d, double* hi2d);

  /**
   * Remove any FACE vertices with one or more unconnected edges
   * unconnected means the edge is only part of this vertex
   * mark vertex as inactive, decrement their edges,
   *   also set edges inactive if they are no longer attached to vertices
   * iterate twice since another face may become unconnected
   */
  void RemoveFaces();

  /**
   * Check BPG for consistency
   * vertices have 3 or more unique edges that point back to it
   * edges have 2 unique vertices
   */
  int Check();

  /**
   * Convert BPG into simple closed polyhdera, not nested
   * walk BPG from any unused vertex, flagging vertices as used
   * stack is list of new vertices to process
   * loop over edges of pgon, add its unused neighbors to stack
   * when stack is empty, loop is closed
   * accumulate volume of polyhedra as walk it from volume of each vertex
   */
  void Walk();

  /**
   * Create Polyhedron from Loop
   */
  int LoopToPolyhedron();

  /**
   * Assign each shell index in list to one of the split cells in PH
   * return surfMap[i] = which PH the Ith shell index is assigned to
   * set surfMap[i] = -1 if the shell did not end up in a PH
   *   could have been discarded in ClipShells()
   *   due to touching cell or lying along a cell edge or face
   */
  void CreateSurfMap();

  /**
   * Find a surf point that is inside or on the boundary of the current cell
   * for external surfs and cells already been flagged as a split cell
   * SurfMap = sub-cell index each surf is part of (-1 if not eligible)
   * return xSplit = coords of point
   * return xSub = sub-cell index the chosen surf is in
   */
  int SplitPoint(vtkIdType cellId, double* lo, double* hi, double* xSplit, int& xSub);

  /**
   * Insert edge iEdge in dir for iVert
   * also update vertex info for added edge.
   */
  void EdgeInsert(vtkIdType iEdge, int dir, vtkIdType iVert,
    vtkIdType iPrev, int dirPrev, vtkIdType iNext, int dirNext);

  /**
   * Complete edge removal in both dirs
   * will leave edge marked inactive
   */
  void EdgeRemove(Edge* edge);

  /**
   * Edge removal in dir
   * also update vertex info for removed edge
   * mark edge inactive if its nVert -> 0
   */
  void EdgeRemove(Edge* edge, int dir);

  /**
   * Remove a vertex and all edges it includes
   */
  void VertexRemove(Vertex* vert);

  /**
   * A planar polygon is grazing if it lies entirely in plane of any face of cell
   * and its normal is outward with respect to cell
   * Return 1 if grazing else 0
   */
  int Grazing(Vertex* vert, double* lo, double* hi);

  /**
   * Identify which cell faces point P is on
   * return list of face IDs (0-5)
   * list length can be 0,1,2
   */
  int OnFaces(double* p, double* lo, double* hi, int* faces);

  /**
   * Identify which cell faces edge between p1,p2 is on
   * p1,p2 assumed to be on surface or interior of cell
   * return list of face IDs (0-5)
   * list length can be 0,1,2
   */
  int WhichFaces(double* p1, double* p2, double* lo, double* hi, int* faces);

  /**
   * Extract 2d cell from iFace (0-5) of 3d cell
   * return lo2d/hi2d = xlo,xhi,ylo,yhi
   * for XLO/XHI, keep (y,z) -> (x,y), look at face from inside/outside 3d cell
   * for YLO/YHI, keep (x,z) -> (x,y), look at face from outside/inside 3d cell
   * for ZLO/ZHI, keep (x,y) -> (x,y), look at face from inside/outside 3d cell
   */
  void FaceFromCell(int iFace, double* lo, double* hi, double* lo2d, double* hi2d);

  /**
   * Compress a 3d point into a 2d point on iFace
   */
  void Compress2D(int iFace, double* p3, double* p2);

  /**
   * Expand a 2d point into 3d point on iFace with extra coord = value
   */
  void Expand2D(int iFace, double value, double* p2, double* p3);

  /**
   * Surtherland-Hodgman clipping algorithm. This doesn't need to
   * don't need to delete duplicate points since touching counts as intersection
   */
  int Clip(vtkIdType shellId, vtkPoints* points,
    const vtkBoundingBox& triBBox, const vtkBoundingBox& cellBBox);

  /**
   * Return intersection pt C of line segment A, B, in dim with coord value
   * guaranteed to interset by caller
   * C can be same as A or B, will just overwite
   */
  void Between(double* a, double* b, int dim, double value, double* c);

  /**
   * Return 1 if x, y are same point, else 0
   */
  int SamePoint(double* x, double* y, double t);

  /**
   * Return 0-7 if pt is a corner pt of grid cell
   * else return -1
   */
  int Corner(double* pt, double* lo, double* hi);

  /**
   * Move point within epsilon of any cell face to be on cell faces
   */
  void MoveToFaces(double* pt, double* lo, double* hi, double epsilon);

  /**
   * Look for edge (x,y) in list of edges which match as (x,y) or (y,x).
   * if flag, do not match edges that are part of a CTRI (style = CTRI,CTRIFACE)
   *   this is used by AddFace() when adding edges of an entire face
   *   this avoids matching an on-face CTRI with norm into cell
   * error if find edge and it is already part of a vertex in that dir
   * return = index if find it, else -1.
   * return dir = 0 if matches as (x,y), 1 if matches as (y,x), -1 if no match
   * return err = 1
   */
  int FindEdge(double* x, double* y, int flag, int& dir);

  /**
   * Calculate cut area of a grid cell that contains nSurf lines
   * also calculate if cell is split into distinct sub-areas by lines
   */
  int SplitFace(int id, int iFace, double* lo, double* hi);

  /**
   * Create Weiler/Atherton data structure within points data struct
   * sets all values for each Point in the data struct
   * check that set of points in clines is valid: 3 possible error returns
   *   no pt should appear no more than twice
   *   if appears twice, must be once as first, once at last
   *   if appears once, must be on cell border
   * then add 4 cell corner pts to points
   *   new CORNER pt if doesn't exist
   *   else already an ENTRY/EXIT pt (not a TWO pt)
   * create linked lists for loops and cell perimeter pts
   */
  int WeilerBuild(double* lo, double* hi);

  /**
   * Create loops data struct with one Loop for each closed path in WA data struct
   * sets all values for each Loop in the data struct
   * discard loop if did not close on itself, b/c just corner pts
   */
  void WeilerLoops(double* lo, double* hi);

  /**
   * Create pgs data struct with one PG (polygon) for each disjoint flow volume
   * a PG can have positive, negative, or zero area
   * some combinations of these are error conditions
   * for no positive areas, return -1 so caller can hande it
   */
  int LoopToPG();

  /**
   * Check if pt is inside or outside or on cell border
   * reutrn EXTERIOR,BORDER,INTERIOR
   */
  int PtFlag(double* pt, double* lo, double* hi);

  /**
   * Check which side of cell pt is on, assumed to be on border
   * not called for corner pt, but return either side
   * return 0,1,2,3 = lower,right,upper,left if on border, -1 if not
   */
  int WhichSide(double* pt, double* lo, double* hi);

  unsigned int Dimension;
  vtkUnstructuredGrid* Output;
  vtkPolyData* Source;
  vtkPoints* Points;
  vtkDataSetAttributes* OutData;

  vtkIncrementalPointLocator* Locator;
  vtkCellLocator* CellLocator;
  vtkImplicitPolyDataDistance* ImplicitFunction;

  vtkTypeBool UseInputBounds;
  double GridBounds[6];
  int GridDimensions[3];
  vtkTypeBool ComputeVolume;
  char* VolumeArrayName;
  vtkTypeBool InsideOut;
  double Tolerance;
  double EpsilonEdge;
  double Shrink;

  vtkIdList* CellSurfs;
  std::unordered_map<vtkIdType, CutCellInfo> CellInfo;
  std::vector<vtkIdType> SurfMap;

  int Empty;
  vtkIdType nVerts;
  vtkIdType nEdges;
  std::vector<Vertex> Verts;
  std::vector<Edge> Edges;
  std::vector<vtkIdType> FaceList[6];

  vtkIdType nPoint2d;
  vtkIdType nLoop2d;
  vtkIdType nLoop3d;
  vtkIdType nPolygons;
  vtkIdType nPolyhedrons;
  std::vector<Cline> CLines; // list of CLines
  std::vector<Point> Points2d; // list of Points = Weiler/Atherthon data structure
  std::vector<Loop2D> Loop2d; // list of loops in Points
  std::vector<Loop3D> Loop3d; // list of loops of vertices = polyhdera
  std::vector<Polygon> Polygons; // list of polygons = one or more loops
  std::vector<Polyhedron> Polyhedons; // list of polyhedrons = one or more loops
  std::vector<int> Used; // 0/1 flag for each point whcn walking loops
  std::vector<vtkIdType> Stack; // list of vertices to check when walking loops
  std::vector<double> Vols;
  vtkIdType GrazeCount; // count of shells that graze cell surf w/ outward norm
  vtkIdType TouchCount; // count of shells that only touch cell surf
  vtkIdType TouchMark; // corner marking inferred by touching shells
  vtkIdType nSurf; // # of surf elements in cell
  double** Path1, ** Path2;

private:
  vtkCutCell3D(const vtkCutCell3D&) = delete;
  void operator=(const vtkCutCell3D&) = delete;
};

VTK_ABI_NAMESPACE_END
#endif